# Copyright 2001 Brad Chapman.
# Revisions copyright 2009-2010 by Peter Cock.
# Revisions copyright 2010 by Phillip Garland.
# All rights reserved.
# This file is part of the Biopython distribution and governed by your
# choice of the "Biopython License Agreement" or the "BSD 3-Clause License".
# Please see the LICENSE file that should have been included as part of this
# package.
"""Definitions for interacting with BLAST related applications.

Wrappers for the new NCBI BLAST+ tools (written in C++):

 - NcbiblastpCommandline - Protein-Protein BLAST
 - NcbiblastnCommandline - Nucleotide-Nucleotide BLAST
 - NcbiblastxCommandline - Translated Query-Protein Subject BLAST
 - NcbitblastnCommandline - Protein Query-Translated Subject BLAST
 - NcbitblastxCommandline - Translated Query-Protein Subject BLAST
 - NcbipsiblastCommandline - Position-Specific Initiated BLAST
 - NcbirpsblastCommandline - Reverse Position Specific BLAST
 - NcbirpstblastnCommandline - Translated Reverse Position Specific BLAST
 - NcbideltablastCommandline - Protein-Protein domain enhanced lookup time accelerated blast
 - NcbiblastformatterCommandline - Convert ASN.1 to other BLAST output formats
 - NcbimakeblastdbCommandline - Application to create BLAST databases

For further details, see:

Camacho et al. BLAST+: architecture and applications
BMC Bioinformatics 2009, 10:421
https://doi.org/10.1186/1471-2105-10-421
"""
from __future__ import print_function

from Bio.Application import _Option, AbstractCommandline, _Switch


class _NcbibaseblastCommandline(AbstractCommandline):
    """Base Commandline object for (new) NCBI BLAST+ wrappers (PRIVATE).

    This is provided for subclassing, it deals with shared options
    common to all the BLAST tools (blastn, rpsblast, rpsblast, etc
    AND blast_formatter).
    """

    def __init__(self, cmd=None, **kwargs):
        assert cmd is not None
        extra_parameters = [
            # Core:
            _Switch(["-h", "h"],
                    "Print USAGE and DESCRIPTION;  ignore other arguments."),
            _Switch(["-help", "help"],
                    "Print USAGE, DESCRIPTION and ARGUMENTS description; "
                    "ignore other arguments."),
            _Switch(["-version", "version"],
                    "Print version number;  ignore other arguments."),
            # Output configuration options
            _Option(["-out", "out"],
                    "Output file for alignment.",
                    filename=True,
                    equate=False),
            # Formatting options:
            _Option(["-outfmt", "outfmt"],
                    "Alignment view.  Typically an integer 0-14 but for some "
                    "formats can be named columns like '6 qseqid sseqid'.  "
                    "Use 5 for XML output (differs from classic BLAST which "
                    "used 7 for XML).",
                    filename=True,  # to ensure spaced inputs are quoted
                    equate=False),
            # TODO - Document and test the column options
            _Switch(["-show_gis", "show_gis"],
                    "Show NCBI GIs in deflines?"),
            _Option(["-num_descriptions", "num_descriptions"],
                    """Number of database sequences to show one-line descriptions for.

                    Integer argument (at least zero). Default is 500.
                    See also num_alignments.""",
                    equate=False),
            _Option(["-num_alignments", "num_alignments"],
                    """Number of database sequences to show num_alignments for.

                    Integer argument (at least zero). Default is 200.
                    See also num_alignments.""",
                    equate=False),
            _Option(["-line_length", "line_length"],
                    """Line length for formatting alignments (integer, at least 1, default 60).

                    Not applicable for outfmt > 4.

                    Added in BLAST+ 2.2.30.
                    """,
                    equate=False),
            _Switch(["-html", "html"],
                    "Produce HTML output? See also the outfmt option."),
            # Miscellaneous options
            _Switch(["-parse_deflines", "parse_deflines"],
                    "Should the query and subject defline(s) be parsed?"),
        ]
        try:
            # Insert extra parameters - at the start just in case there
            # are any arguments which must come last:
            self.parameters = extra_parameters + self.parameters
        except AttributeError:
            # Should we raise an error?  The subclass should have set this up!
            self.parameters = extra_parameters
        AbstractCommandline.__init__(self, cmd, **kwargs)

    def _validate_incompatibilities(self, incompatibles):
        """Validate parameters for incompatibilities (PRIVATE).

        Used by the _validate method.
        """
        for a in incompatibles:
            if self._get_parameter(a):
                for b in incompatibles[a]:
                    if self._get_parameter(b):
                        raise ValueError("Options %s and %s are incompatible."
                                         % (a, b))


class _NcbiblastCommandline(_NcbibaseblastCommandline):
    """Base Commandline object for (new) NCBI BLAST+ wrappers (PRIVATE).

    This is provided for subclassing, it deals with shared options
    common to all the BLAST tools (blastn, rpsblast, rpsblast, etc).
    """

    def __init__(self, cmd=None, **kwargs):
        assert cmd is not None
        extra_parameters = [
            # Input query options:
            _Option(["-query", "query"],
                    "The sequence to search with.",
                    filename=True,
                    equate=False),  # Should this be required?
            _Option(["-query_loc", "query_loc"],
                    "Location on the query sequence (Format: start-stop)",
                    equate=False),
            # General search options:
            _Option(["-db", "db"],
                    "The database to BLAST against.",
                    equate=False),
            _Option(["-evalue", "evalue"],
                    "Expectation value cutoff.",
                    equate=False),
            _Option(["-word_size", "word_size"],
                    """Word size for wordfinder algorithm.

                    Integer. Minimum 2.""",
                    equate=False),
            # BLAST-2-Sequences options:
            # - see subclass
            # Formatting options:
            # - see baseclass
            # Query filtering options
            _Option(["-soft_masking", "soft_masking"],
                    "Apply filtering locations as soft masks (Boolean, Default = true)",
                    equate=False),
            _Switch(["-lcase_masking", "lcase_masking"],
                    "Use lower case filtering in query and subject sequence(s)?"),
            # Restrict search or results
            _Option(["-gilist", "gilist"],
                    """Restrict search of database to list of GI's.

                    Incompatible with: negative_gilist, seqidlist, negative_seqidlist,
                    remote, subject, subject_loc""",
                    filename=True,
                    equate=False),
            _Option(["-negative_gilist", "negative_gilist"],
                    """Restrict search of database to everything except the listed GIs.

                    Incompatible with: gilist, seqidlist, remote, subject, subject_loc""",
                    filename=True,
                    equate=False),
            _Option(["-seqidlist", "seqidlist"],
                    """Restrict search of database to list of SeqID's.

                    Incompatible with: gilist, negative_gilist, remote, subject, subject_loc""",
                    filename=True,
                    equate=False),
            _Option(["-negative_seqidlist", "negative_seqidlist"],
                    """Restrict search of database to everything except the listed SeqID's.

                    Incompatible with: gilist, seqidlist, remote, subject, subject_loc""",
                    filename=True,
                    equate=False),
            _Option(["-entrez_query", "entrez_query"],
                    "Restrict search with the given Entrez query (requires remote).",
                    equate=False),
            _Option(["-qcov_hsp_perc", "qcov_hsp_perc"],
                    """Percent query coverage per hsp (float, 0 to 100).

                    Added in BLAST+ 2.2.30.
                    """,
                    equate=False),
            _Option(["-max_target_seqs", "max_target_seqs"],
                    "Maximum number of aligned sequences to keep (integer, at least one).",
                    equate=False),
            # Statistical options
            _Option(["-dbsize", "dbsize"],
                    "Effective length of the database (integer).",
                    equate=False),
            _Option(["-searchsp", "searchsp"],
                    "Effective length of the search space (integer).",
                    equate=False),
            _Option(["-max_hsps_per_subject", "max_hsps_per_subject"],
                    "Override max number of HSPs per subject saved for ungapped searches (integer).",
                    equate=False),
            _Option(["-max_hsps", "max_hsps"],
                    "Set max number of HSPs saved per subject sequence (default 0 means no limit).",
                    equate=False),
            _Switch(["-sum_statistics", "sum_statistics"],
                    "Use sum statistics."),
            # Is -sum_stats a BLAST+ bug, why not use -sum_statistics switch?
            _Option(["-sum_stats", "sum_stats"],
                    """Use sum statistics (boolean).

                    Added in BLAST+ 2.2.30.
                    """,
                    equate=False),
            # Extension options
            _Option(["-xdrop_ungap", "xdrop_ungap"],
                    "X-dropoff value (in bits) for ungapped extensions (float).",
                    equate=False),
            _Option(["-xdrop_gap", "xdrop_gap"],
                    "X-dropoff value (in bits) for preliminary gapped extensions (float).",
                    equate=False),
            _Option(["-xdrop_gap_final", "xdrop_gap_final"],
                    "X-dropoff value (in bits) for final gapped alignment (float).",
                    equate=False),
            _Option(["-window_size", "window_size"],
                    "Multiple hits window size, use 0 to specify 1-hit algorithm (integer).",
                    equate=False),
            # Search strategy options
            _Option(["-import_search_strategy", "import_search_strategy"],
                    """Search strategy to use.

                    Incompatible with: export_search_strategy""",
                    filename=True,
                    equate=False),
            _Option(["-export_search_strategy", "export_search_strategy"],
                    """File name to record the search strategy used.

                    Incompatible with: import_search_strategy""",
                    filename=True,
                    equate=False),
            # Miscellaneous options
            _Option(["-num_threads", "num_threads"],
                    """Number of threads to use in the BLAST search (integer, at least one).

                    Default is one.
                    Incompatible with: remote""",
                    equate=False),
            _Switch(["-remote", "remote"],
                    """Execute search remotely?

                    Incompatible with: gilist, negative_gilist, subject_loc, num_threads, ..."""),
        ]
        try:
            # Insert extra parameters - at the start just in case there
            # are any arguments which must come last:
            self.parameters = extra_parameters + self.parameters
        except AttributeError:
            # Should we raise an error?  The subclass should have set this up!
            self.parameters = extra_parameters
        _NcbibaseblastCommandline.__init__(self, cmd, **kwargs)

    def _validate(self):
        incompatibles = {"remote": ["gilist", "negative_gilist", "num_threads"],
                         "import_search_strategy": ["export_search_strategy"],
                         "gilist": ["negative_gilist"],
                         "seqidlist": ["gilist", "negative_gilist", "remote"]}
        self._validate_incompatibilities(incompatibles)
        if self.entrez_query and not self.remote:
            raise ValueError("Option entrez_query requires remote option.")
        AbstractCommandline._validate(self)


class _Ncbiblast2SeqCommandline(_NcbiblastCommandline):
    """Base Commandline object for (new) NCBI BLAST+ wrappers (PRIVATE).

    This is provided for subclassing, it deals with shared options
    common to all the BLAST tools supporting two-sequence BLAST
    (blastn, psiblast, etc) but not rpsblast or rpstblastn.
    """

    def __init__(self, cmd=None, **kwargs):
        assert cmd is not None
        extra_parameters = [
            # General search options:
            _Option(["-gapopen", "gapopen"],
                    "Cost to open a gap (integer).",
                    equate=False),
            _Option(["-gapextend", "gapextend"],
                    "Cost to extend a gap (integer).",
                    equate=False),
            # BLAST-2-Sequences options:
            _Option(["-subject", "subject"],
                    """Subject sequence(s) to search.

                    Incompatible with: db, gilist, seqidlist, negative_gilist,
                    negative_seqidlist, db_soft_mask, db_hard_mask

                    See also subject_loc.""",
                    filename=True,
                    equate=False),
            _Option(["-subject_loc", "subject_loc"],
                    """Location on the subject sequence (Format: start-stop).

                    Incompatible with: db, gilist, seqidlist, negative_gilist,
                    negative_seqidlist, db_soft_mask, db_hard_mask, remote.

                    See also subject.""",
                    equate=False),
            # Restrict search or results:
            _Option(["-culling_limit", "culling_limit"],
                    """Hit culling limit (integer).

                    If the query range of a hit is enveloped by that of at
                    least this many higher-scoring hits, delete the hit.

                    Incompatible with: best_hit_overhang, best_hit_score_edge.
                    """,
                    equate=False),
            _Option(["-best_hit_overhang", "best_hit_overhang"],
                    """Best Hit algorithm overhang value (float, recommended value: 0.1)

                    Float between 0.0 and 0.5 inclusive.

                    Incompatible with: culling_limit.""",
                    equate=False),
            _Option(["-best_hit_score_edge", "best_hit_score_edge"],
                    """Best Hit algorithm score edge value (float, recommended value: 0.1)

                    Float between 0.0 and 0.5 inclusive.

                    Incompatible with: culling_limit.""",
                    equate=False),
        ]
        try:
            # Insert extra parameters - at the start just in case there
            # are any arguments which must come last:
            self.parameters = extra_parameters + self.parameters
        except AttributeError:
            # Should we raise an error?  The subclass should have set this up!
            self.parameters = extra_parameters
        _NcbiblastCommandline.__init__(self, cmd, **kwargs)

    def _validate(self):
        incompatibles = {"subject_loc": ["db", "gilist", "negative_gilist", "seqidlist", "remote"],
                         "culling_limit": ["best_hit_overhang", "best_hit_score_edge"],
                         "subject": ["db", "gilist", "negative_gilist", "seqidlist"]}
        self._validate_incompatibilities(incompatibles)
        _NcbiblastCommandline._validate(self)


class _NcbiblastMain2SeqCommandline(_Ncbiblast2SeqCommandline):
    """Base Commandline object for (new) NCBI BLAST+ wrappers (PRIVATE).

    This is provided for subclassing, it deals with shared options
    common to the main BLAST tools blastp, blastn, blastx, tblastx, tblastn
    but not psiblast, rpsblast or rpstblastn.
    """

    def __init__(self, cmd=None, **kwargs):
        assert cmd is not None
        extra_parameters = [
            # Restrict search or results:
            _Option(["-db_soft_mask", "db_soft_mask"],
                    """Filtering algorithm for soft masking (integer).

                    Filtering algorithm ID to apply to the BLAST database as soft masking.

                    Incompatible with: db_hard_mask, subject, subject_loc""",
                    equate=False),
            _Option(["-db_hard_mask", "db_hard_mask"],
                    """Filtering algorithm for hard masking (integer).

                    Filtering algorithm ID to apply to the BLAST database as hard masking.

                    Incompatible with: db_soft_mask, subject, subject_loc""",
                    equate=False),
        ]
        try:
            # Insert extra parameters - at the start just in case there
            # are any arguments which must come last:
            self.parameters = extra_parameters + self.parameters
        except AttributeError:
            # Should we raise an error?  The subclass should have set this up!
            self.parameters = extra_parameters
        _Ncbiblast2SeqCommandline.__init__(self, cmd, **kwargs)

    def _validate(self):
        incompatibles = {"db_soft_mask": ["db_hard_mask", "subject", "subject_loc"],
                         "db_hard_mask": ["db_soft_mask", "subject", "subject_loc"]}
        self._validate_incompatibilities(incompatibles)
        _Ncbiblast2SeqCommandline._validate(self)


class NcbiblastpCommandline(_NcbiblastMain2SeqCommandline):
    """Create a commandline for the NCBI BLAST+ program blastp (for proteins).

    With the release of BLAST+ (BLAST rewritten in C++ instead of C), the NCBI
    replaced the old blastall tool with separate tools for each of the searches.
    This wrapper therefore replaces BlastallCommandline with option -p blastp.

    >>> from Bio.Blast.Applications import NcbiblastpCommandline
    >>> cline = NcbiblastpCommandline(query="rosemary.pro", db="nr",
    ...                               evalue=0.001, remote=True, ungapped=True)
    >>> cline
    NcbiblastpCommandline(cmd='blastp', query='rosemary.pro', db='nr', evalue=0.001, remote=True, ungapped=True)
    >>> print(cline)
    blastp -query rosemary.pro -db nr -evalue 0.001 -remote -ungapped

    You would typically run the command line with cline() or via the Python
    subprocess module, as described in the Biopython tutorial.
    """

    def __init__(self, cmd="blastp", **kwargs):
        """Initialize the class."""
        self.parameters = [
            # General search options:
            _Option(["-task", "task"],
                    "Task to execute (string, blastp (default), blastp-fast or blastp-short).",
                    checker_function=lambda value: value in ["blastp",
                                                             "blastp-fast",
                                                             "blastp-short"],
                    equate=False),
            _Option(["-matrix", "matrix"],
                    "Scoring matrix name (default BLOSUM62)."),
            _Option(["-threshold", "threshold"],
                    "Minimum score for words to be added to the BLAST lookup table (float).",
                    equate=False),
            _Option(["-comp_based_stats", "comp_based_stats"],
                    """Use composition-based statistics (string, default 2, i.e. True).

                    0, F or f: no composition-based statistics
                    2, T or t, D or d : Composition-based score adjustment as in
                    Bioinformatics 21:902-911, 2005, conditioned on sequence properties

                    Note that tblastn also supports values of 1 and 3.""",
                    checker_function=lambda value: value in "0Ft2TtDd",
                    equate=False),
            # Query filtering options:
            _Option(["-seg", "seg"],
                    """Filter query sequence with SEG (string).

                    Format: "yes", "window locut hicut", or "no" to disable.
                    Default is "12 2.2 2.5""",
                    equate=False),
            # Extension options:
            _Switch(["-ungapped", "ungapped"],
                    "Perform ungapped alignment only?"),
            # Miscellaneous options:
            _Switch(["-use_sw_tback", "use_sw_tback"],
                    "Compute locally optimal Smith-Waterman alignments?"),
        ]
        _NcbiblastMain2SeqCommandline.__init__(self, cmd, **kwargs)


class NcbiblastnCommandline(_NcbiblastMain2SeqCommandline):
    """Wrapper for the NCBI BLAST+ program blastn (for nucleotides).

    With the release of BLAST+ (BLAST rewritten in C++ instead of C), the NCBI
    replaced the old blastall tool with separate tools for each of the searches.
    This wrapper therefore replaces BlastallCommandline with option -p blastn.

    For example, to run a search against the "nt" nucleotide database using the
    FASTA nucleotide file "m_code.fasta" as the query, with an expectation value
    cut off of 0.001, saving the output to a file in XML format:

    >>> from Bio.Blast.Applications import NcbiblastnCommandline
    >>> cline = NcbiblastnCommandline(query="m_cold.fasta", db="nt", strand="plus",
    ...                               evalue=0.001, out="m_cold.xml", outfmt=5)
    >>> cline
    NcbiblastnCommandline(cmd='blastn', out='m_cold.xml', outfmt=5, query='m_cold.fasta', db='nt', evalue=0.001, strand='plus')
    >>> print(cline)
    blastn -out m_cold.xml -outfmt 5 -query m_cold.fasta -db nt -evalue 0.001 -strand plus

    You would typically run the command line with cline() or via the Python
    subprocess module, as described in the Biopython tutorial.
    """

    def __init__(self, cmd="blastn", **kwargs):
        """Initialize the class."""
        self.parameters = [
            # Input query options:
            _Option(["-strand", "strand"],
                    """Query strand(s) to search against database/subject.

                    Values allowed are "both" (default), "minus", "plus".""",
                    checker_function=lambda value: value in ["both",
                                                             "minus",
                                                             "plus"],
                    equate=False),
            # General search options:
            _Option(["-task", "task"],
                    """Task to execute (string, default 'megablast')

                    Allowed values 'blastn', 'blastn-short', 'dc-megablast', 'megablast'
                    (the default), or 'vecscreen'.""",
                    checker_function=lambda value: value in ["blastn",
                                                             "blastn-short",
                                                             "dc-megablast",
                                                             "megablast",
                                                             "vecscreen"],
                    equate=False),
            _Option(["-penalty", "penalty"],
                    "Penalty for a nucleotide mismatch (integer, at most zero).",
                    equate=False),
            _Option(["-reward", "reward"],
                    "Reward for a nucleotide match (integer, at least zero).",
                    equate=False),
            _Option(["-use_index", "use_index"],
                    "Use MegaBLAST database index (Boolean, Default = False)",
                    equate=False),
            _Option(["-index_name", "index_name"],
                    "MegaBLAST database index name.",
                    equate=False),
            # Query filtering options:
            _Option(["-dust", "dust"],
                    """Filter query sequence with DUST (string).

                    Format: 'yes', 'level window linker', or 'no' to disable.
                    Default = '20 64 1'.
                    """,
                    equate=False),
            _Option(["-filtering_db", "filtering_db"],
                    "BLAST database containing filtering elements (i.e. repeats).",
                    equate=False),
            _Option(["-window_masker_taxid", "window_masker_taxid"],
                    "Enable WindowMasker filtering using a Taxonomic ID (integer).",
                    equate=False),
            _Option(["-window_masker_db", "window_masker_db"],
                    "Enable WindowMasker filtering using this repeats database (string).",
                    equate=False),
            # Restrict search or results:
            _Option(["-perc_identity", "perc_identity"],
                    "Percent identity (real, 0 to 100 inclusive).",
                    equate=False),
            # Discontiguous MegaBLAST options
            _Option(["-template_type", "template_type"],
                    """Discontiguous MegaBLAST template type (string).

                    Allowed values: 'coding', 'coding_and_optimal' or 'optimal'
                    Requires: template_length.""",
                    checker_function=lambda value: value in ["coding", "coding_and_optimal", "optimal"],
                    equate=False),
            _Option(["-template_length", "template_length"],
                    """Discontiguous MegaBLAST template length (integer).

                    Allowed values: 16, 18, 21

                    Requires: template_type.""",
                    checker_function=lambda value: value in [16, 18, 21, "16", "18", "21"],
                    equate=False),
            # Extension options:
            _Switch(["-no_greedy", "no_greedy"],
                    "Use non-greedy dynamic programming extension"),
            _Option(["-min_raw_gapped_score", "min_raw_gapped_score"],
                    "Minimum raw gapped score to keep an alignment in the "
                    "preliminary gapped and traceback stages (integer).",
                    equate=False),
            _Switch(["-ungapped", "ungapped"],
                    "Perform ungapped alignment only?"),
            _Option(["-off_diagonal_range", "off_diagonal_range"],
                    """Number of off-diagonals to search for the 2nd hit (integer).

                    Expects a positive integer, or 0 (default) to turn off.

                    Added in BLAST 2.2.23+
                    """,
                    equate=False),
        ]
        _NcbiblastMain2SeqCommandline.__init__(self, cmd, **kwargs)

    def _validate(self):
        if (self.template_type and not self.template_length) \
                or (self.template_length and not self.template_type):
            raise ValueError("Options template_type and template_type require each other.")
        _NcbiblastMain2SeqCommandline._validate(self)


class NcbiblastxCommandline(_NcbiblastMain2SeqCommandline):
    """Wrapper for the NCBI BLAST+ program blastx (nucleotide query, protein database).

    With the release of BLAST+ (BLAST rewritten in C++ instead of C), the NCBI
    replaced the old blastall tool with separate tools for each of the searches.
    This wrapper therefore replaces BlastallCommandline with option -p blastx.

    >>> from Bio.Blast.Applications import NcbiblastxCommandline
    >>> cline = NcbiblastxCommandline(query="m_cold.fasta", db="nr", evalue=0.001)
    >>> cline
    NcbiblastxCommandline(cmd='blastx', query='m_cold.fasta', db='nr', evalue=0.001)
    >>> print(cline)
    blastx -query m_cold.fasta -db nr -evalue 0.001

    You would typically run the command line with cline() or via the Python
    subprocess module, as described in the Biopython tutorial.
    """

    def __init__(self, cmd="blastx", **kwargs):
        """Initialize the class."""
        self.parameters = [
            # Input query options:
            _Option(["-task", "task"],
                    "Task to execute (string, blastx (default) or blastx-fast).",
                    checker_function=lambda value: value in ["blastx",
                                                             "blastx-fast"],
                    equate=False),
            _Option(["-strand", "strand"],
                    """Query strand(s) to search against database/subject.

                    Values allowed are "both" (default), "minus", "plus".""",
                    checker_function=lambda value: value in ["both", "minus", "plus"],
                    equate=False),
            # Input query options:
            _Option(["-query_gencode", "query_gencode"],
                    "Genetic code to use to translate query (integer, default 1).",
                    equate=False),
            # General search options:
            _Option(["-frame_shift_penalty", "frame_shift_penalty"],
                    """Frame shift penalty (integer, at least 1, default ignored) (OBSOLETE).

                    This was removed in BLAST 2.2.27+""",
                    equate=False),
            _Option(["-max_intron_length", "max_intron_length"],
                    """Maximum intron length (integer).

                    Length of the largest intron allowed in a translated nucleotide
                    sequence when linking multiple distinct alignments (a negative
                    value disables linking). Default zero.""",
                    equate=False),
            _Option(["-matrix", "matrix"],
                    "Scoring matrix name (default BLOSUM62).",
                    equate=False),
            _Option(["-threshold", "threshold"],
                    "Minimum score for words to be added to the BLAST lookup table (float).",
                    equate=False),
            _Option(["-comp_based_stats", "comp_based_stats"],
                    """Use composition-based statistics for blastp, blastx, or tblastn:

                    D or d: default (equivalent to 2 )

                    0 or F or f: no composition-based statistics

                    1: Composition-based statistics as in NAR 29:2994-3005, 2001

                    2 or T or t : Composition-based score adjustment as in
                    Bioinformatics 21:902-911, 2005, conditioned on sequence properties

                    3: Composition-based score adjustment as in
                    Bioinformatics 21:902-911, 2005, unconditionally

                    For programs other than tblastn, must either be absent or be D, F or 0
                    Default = 2.
                    """,
                    equate=False),
            # Query filtering options:
            _Option(["-seg", "seg"],
                    """Filter query sequence with SEG (string).

                    Format: "yes", "window locut hicut", or "no" to disable.
                    Default is "12 2.2 2.5""",
                    equate=False),
            # Extension options:
            _Switch(["-ungapped", "ungapped"],
                    "Perform ungapped alignment only?"),
            _Switch(["-use_sw_tback", "use_sw_tback"],
                    "Compute locally optimal Smith-Waterman alignments?"),
        ]
        _NcbiblastMain2SeqCommandline.__init__(self, cmd, **kwargs)


class NcbitblastnCommandline(_NcbiblastMain2SeqCommandline):
    """Wrapper for the NCBI BLAST+ program tblastn.

    With the release of BLAST+ (BLAST rewritten in C++ instead of C), the NCBI
    replaced the old blastall tool with separate tools for each of the searches.
    This wrapper therefore replaces BlastallCommandline with option -p tblastn.

    >>> from Bio.Blast.Applications import NcbitblastnCommandline
    >>> cline = NcbitblastnCommandline(help=True)
    >>> cline
    NcbitblastnCommandline(cmd='tblastn', help=True)
    >>> print(cline)
    tblastn -help

    You would typically run the command line with cline() or via the Python
    subprocess module, as described in the Biopython tutorial.
    """

    def __init__(self, cmd="tblastn", **kwargs):
        """Initialize the class."""
        self.parameters = [
            # General search options:
            _Option(["-task", "task"],
                    "Task to execute (string, tblastn (default) or tblastn-fast).",
                    checker_function=lambda value: value in ["tblastn",
                                                             "tblastn-fast"],
                    equate=False),
            _Option(["-db_gencode", "db_gencode"],
                    "Genetic code to use to translate query (integer, default 1).",
                    equate=False),
            _Option(["-frame_shift_penalty", "frame_shift_penalty"],
                    """Frame shift penalty (integer, at least 1, default ignored) (OBSOLETE).

                    This was removed in BLAST 2.2.27+""",
                    equate=False),
            _Option(["-max_intron_length", "max_intron_length"],
                    """Maximum intron length (integer).

                    Length of the largest intron allowed in a translated nucleotide
                    sequence when linking multiple distinct alignments (a negative
                    value disables linking). Default zero.""",
                    equate=False),
            _Option(["-matrix", "matrix"],
                    "Scoring matrix name (default BLOSUM62).",
                    equate=False),
            _Option(["-threshold", "threshold"],
                    "Minimum score for words to be added to the BLAST lookup table (float).",
                    equate=False),
            _Option(["-comp_based_stats", "comp_based_stats"],
                    r"""Use composition-based statistics (string, default 2, i.e. True).

                    0, F or f: no composition-based statistics

                    1: Composition-based statistics as in NAR 29:2994-3005, 2001

                    2, T or t, D or d : Composition-based score adjustment as in
                    Bioinformatics 21:902-911, 2005, conditioned on sequence properties

                    3: Composition-based score adjustment as in Bioinformatics 21:902-911,
                    2005, unconditionally

                    Note that only tblastn supports values of 1 and 3.""",
                    checker_function=lambda value: value in "0Ft12TtDd3",
                    equate=False),
            # Query filtering options:
            _Option(["-seg", "seg"],
                    """Filter query sequence with SEG (string).

                    Format: "yes", "window locut hicut", or "no" to disable.

                    Default is "12 2.2 2.5"
                    """,
                    equate=False),
            # Extension options:
            _Switch(["-ungapped", "ungapped"],
                    "Perform ungapped alignment only?"),
            # Miscellaneous options:
            _Switch(["-use_sw_tback", "use_sw_tback"],
                    "Compute locally optimal Smith-Waterman alignments?"),
            # PSI-TBLASTN options:
            _Option(["-in_pssm", "in_pssm"],
                    """PSI-BLAST checkpoint file

                    Incompatible with: remote, query""",
                    filename=True,
                    equate=False),
        ]
        _NcbiblastMain2SeqCommandline.__init__(self, cmd, **kwargs)


class NcbitblastxCommandline(_NcbiblastMain2SeqCommandline):
    """Wrapper for the NCBI BLAST+ program tblastx.

    With the release of BLAST+ (BLAST rewritten in C++ instead of C), the NCBI
    replaced the old blastall tool with separate tools for each of the searches.
    This wrapper therefore replaces BlastallCommandline with option -p tblastx.

    >>> from Bio.Blast.Applications import NcbitblastxCommandline
    >>> cline = NcbitblastxCommandline(help=True)
    >>> cline
    NcbitblastxCommandline(cmd='tblastx', help=True)
    >>> print(cline)
    tblastx -help

    You would typically run the command line with cline() or via the Python
    subprocess module, as described in the Biopython tutorial.
    """

    def __init__(self, cmd="tblastx", **kwargs):
        """Initialize the class."""
        self.parameters = [
            # Input query options:
            _Option(["-strand", "strand"],
                    """Query strand(s) to search against database/subject.

                    Values allowed are "both" (default), "minus", "plus".""",
                    checker_function=lambda value: value in ["both", "minus", "plus"],
                    equate=False),
            # Input query options:
            _Option(["-query_gencode", "query_gencode"],
                    "Genetic code to use to translate query (integer, default 1).",
                    equate=False),
            # General search options:
            _Option(["-db_gencode", "db_gencode"],
                    "Genetic code to use to translate query (integer, default 1).",
                    equate=False),
            _Option(["-max_intron_length", "max_intron_length"],
                    """Maximum intron length (integer).

                    Length of the largest intron allowed in a translated nucleotide
                    sequence when linking multiple distinct alignments (a negative
                    value disables linking). Default zero.""",
                    equate=False),
            _Option(["-matrix", "matrix"],
                    "Scoring matrix name (default BLOSUM62).",
                    equate=False),
            _Option(["-threshold", "threshold"],
                    "Minimum score for words to be added to the BLAST lookup table (float).",
                    equate=False),
            # Query filtering options:
            _Option(["-seg", "seg"],
                    """Filter query sequence with SEG (string).

                    Format: "yes", "window locut hicut", or "no" to disable.
                    Default is "12 2.2 2.5"
                    """,
                    equate=False),
        ]
        _NcbiblastMain2SeqCommandline.__init__(self, cmd, **kwargs)


class NcbipsiblastCommandline(_Ncbiblast2SeqCommandline):
    """Wrapper for the NCBI BLAST+ program psiblast.

    With the release of BLAST+ (BLAST rewritten in C++ instead of C), the NCBI
    replaced the old blastpgp tool with a similar tool psiblast. This wrapper
    therefore replaces BlastpgpCommandline, the wrapper for blastpgp.

    >>> from Bio.Blast.Applications import NcbipsiblastCommandline
    >>> cline = NcbipsiblastCommandline(help=True)
    >>> cline
    NcbipsiblastCommandline(cmd='psiblast', help=True)
    >>> print(cline)
    psiblast -help

    You would typically run the command line with cline() or via the Python
    subprocess module, as described in the Biopython tutorial.
    """

    def __init__(self, cmd="psiblast", **kwargs):
        """Initialize the class."""
        self.parameters = [
            # General search options:
            _Option(["-matrix", "matrix"],
                    "Scoring matrix name (default BLOSUM62).",
                    equate=False),
            _Option(["-threshold", "threshold"],
                    "Minimum score for words to be added to the BLAST lookup table (float).",
                    equate=False),
            _Option(["-comp_based_stats", "comp_based_stats"],
                    """Use composition-based statistics (string, default 2, i.e. True).

                    0, F or f: no composition-based statistics
                    2, T or t, D or d : Composition-based score adjustment
                    as in Bioinformatics 21:902-911, 2005, conditioned on
                    sequence properties

                    Note that tblastn also supports values of 1 and 3.""",
                    checker_function=lambda value: value in "0Ft2TtDd",
                    equate=False),
            # Query filtering options:
            _Option(["-seg", "seg"],
                    """Filter query sequence with SEG (string).

                    Format: "yes", "window locut hicut", or "no" to disable.
                    Default is "12 2.2 2.5""",
                    equate=False),
            # Extension options:
            _Option(["-gap_trigger", "gap_trigger"],
                    "Number of bits to trigger gapping (float, default 22).",
                    equate=False),
            # Miscellaneous options:
            _Switch(["-use_sw_tback", "use_sw_tback"],
                    "Compute locally optimal Smith-Waterman alignments?"),
            # PSI-BLAST options:
            _Option(["-num_iterations", "num_iterations"],
                    """Number of iterations to perform (integer, at least one).

                    Default is one.
                    Incompatible with: remote""",
                    equate=False),
            _Option(["-out_pssm", "out_pssm"],
                    "File name to store checkpoint file.",
                    filename=True,
                    equate=False),
            _Option(["-out_ascii_pssm", "out_ascii_pssm"],
                    "File name to store ASCII version of PSSM.",
                    filename=True,
                    equate=False),
            _Switch(["-save_pssm_after_last_round", "save_pssm_after_last_round"],
                    "Save PSSM after the last database search."),
            _Switch(["-save_each_pssm", "save_each_pssm"],
                    """Save PSSM after each iteration

                    File name is given in -save_pssm or -save_ascii_pssm options.
                    """),
            _Option(["-in_msa", "in_msa"],
                    """File name of multiple sequence alignment to restart PSI-BLAST.

                    Incompatible with: in_pssm, query""",
                    filename=True,
                    equate=False),
            _Option(["-msa_master_idx", "msa_master_idx"],
                    """Index of sequence to use as master in MSA.

                    Index (1-based) of sequence to use as the master in the
                    multiple sequence alignment. If not specified, the first
                    sequence is used.""",
                    equate=False),
            _Option(["-in_pssm", "in_pssm"],
                    """PSI-BLAST checkpoint file.

                    Incompatible with: in_msa, query, phi_pattern""",
                    filename=True,
                    equate=False),
            # PSSM engine options:
            _Option(["-pseudocount", "pseudocount"],
                    """Pseudo-count value used when constructing PSSM.

                    Integer. Default is zero.""",
                    equate=False),
            _Option(["-inclusion_ethresh", "inclusion_ethresh"],
                    "E-value inclusion threshold for pairwise alignments (float, default 0.002).",
                    equate=False),
            _Switch(["-ignore_msa_master", "ignore_msa_master"],
                    """Ignore the master sequence when creating PSSM

                    Requires: in_msa
                    Incompatible with: msa_master_idx, in_pssm, query, query_loc, phi_pattern
                    """),
            # PHI-BLAST options:
            _Option(["-phi_pattern", "phi_pattern"],
                    """File name containing pattern to search.

                    Incompatible with: in_pssm""",
                    filename=True,
                    equate=False),
        ]
        _Ncbiblast2SeqCommandline.__init__(self, cmd, **kwargs)

    def _validate(self):
        incompatibles = {"num_iterations": ["remote"],
                         "in_msa": ["in_pssm", "query"],
                         "in_pssm": ["in_msa", "query", "phi_pattern"],
                         "ignore_msa_master": ["msa_master_idx", "in_pssm",
                                               "query", "query_loc", "phi_pattern"],
                         }
        self._validate_incompatibilities(incompatibles)
        _Ncbiblast2SeqCommandline._validate(self)


class NcbirpsblastCommandline(_NcbiblastCommandline):
    """Wrapper for the NCBI BLAST+ program rpsblast.

    With the release of BLAST+ (BLAST rewritten in C++ instead of C), the NCBI
    replaced the old rpsblast tool with a similar tool of the same name. This
    wrapper replaces RpsBlastCommandline, the wrapper for the old rpsblast.

    >>> from Bio.Blast.Applications import NcbirpsblastCommandline
    >>> cline = NcbirpsblastCommandline(help=True)
    >>> cline
    NcbirpsblastCommandline(cmd='rpsblast', help=True)
    >>> print(cline)
    rpsblast -help

    You would typically run the command line with cline() or via the Python
    subprocess module, as described in the Biopython tutorial.
    """

    def __init__(self, cmd="rpsblast", **kwargs):
        """Initialize the class."""
        # TODO - remove the -word_size argument as per BLAST+ 2.2.30
        # (BLAST team say it should never have been included, since
        # the word size is set when building the domain database.)
        # This likely means reviewing the class hierarchy again.
        self.parameters = [
            # Query filtering options:
            _Option(["-seg", "seg"],
                    r"""Filter query sequence with SEG (string).

                    Format: "yes", "window locut hicut", or "no" to disable.
                    Default is "12 2.2 2.5"
                    """,
                    equate=False),
            # Restrict search or results:
            _Option(["-culling_limit", "culling_limit"],
                    r"""Hit culling limit (integer).

                    If the query range of a hit is enveloped by that of at
                    least this many higher-scoring hits, delete the hit.

                    Incompatible with: best_hit_overhang, best_hit_score_edge.
                    """,
                    equate=False),
            _Option(["-best_hit_overhang", "best_hit_overhang"],
                    r"""Best Hit algorithm overhang value (recommended value: 0.1)

                    Float between 0.0 and 0.5 inclusive.

                    Incompatible with: culling_limit.""",
                    equate=False),
            _Option(["-best_hit_score_edge", "best_hit_score_edge"],
                    r"""Best Hit algorithm score edge value (recommended value: 0.1)

                    Float between 0.0 and 0.5 inclusive.

                    Incompatible with: culling_limit.""",
                    equate=False),
            # General search options:
            _Option(["-comp_based_stats", "comp_based_stats"],
                    r"""Use composition-based statistics.

                    D or d: default (equivalent to 0 )

                    0 or F or f: Simplified Composition-based statistics as in
                    Bioinformatics 15:1000-1011, 1999

                    1 or T or t: Composition-based statistics as in NAR 29:2994-3005, 2001

                    Default = 0.
                    """,
                    checker_function=lambda value: value in "Dd0Ff1Tt",
                    equate=False),
            # Misc options:
            _Switch(["-use_sw_tback", "use_sw_tback"],
                    r"Compute locally optimal Smith-Waterman alignments?"),
        ]
        _NcbiblastCommandline.__init__(self, cmd, **kwargs)

    def _validate(self):
        incompatibles = {"culling_limit": ["best_hit_overhang", "best_hit_score_edge"]}
        self._validate_incompatibilities(incompatibles)
        _NcbiblastCommandline._validate(self)


class NcbirpstblastnCommandline(_NcbiblastCommandline):
    """Wrapper for the NCBI BLAST+ program rpstblastn.

    With the release of BLAST+ (BLAST rewritten in C++ instead of C), the NCBI
    replaced the old rpsblast tool with a similar tool of the same name, and a
    separate tool rpstblastn for Translated Reverse Position Specific BLAST.

    >>> from Bio.Blast.Applications import NcbirpstblastnCommandline
    >>> cline = NcbirpstblastnCommandline(help=True)
    >>> cline
    NcbirpstblastnCommandline(cmd='rpstblastn', help=True)
    >>> print(cline)
    rpstblastn -help

    You would typically run the command line with cline() or via the Python
    subprocess module, as described in the Biopython tutorial.
    """

    def __init__(self, cmd="rpstblastn", **kwargs):
        """Initialize the class."""
        # TODO - remove the -word_size argument as per BLAST+ 2.2.30
        # (BLAST team say it should never have been included, since
        # the word size is set when building the domain database.)
        # This likely means reviewing the class hierarchy again.
        self.parameters = [
            # Input query options:
            _Option(["-strand", "strand"],
                    r"""Query strand(s) to search against database/subject.

                    Values allowed are "both" (default), "minus", "plus".""",
                    checker_function=lambda value: value in ["both",
                                                             "minus",
                                                             "plus"],
                    equate=False),
            # Input query options:
            _Option(["-query_gencode", "query_gencode"],
                    r"Genetic code to use to translate query (integer, default 1).",
                    equate=False),
            # Query filtering options:
            _Option(["-seg", "seg"],
                    r"""Filter query sequence with SEG (string).

                    Format: "yes", "window locut hicut", or "no" to disable.
                    Default is "12 2.2 2.5" """,
                    equate=False),
            # General search options:
            _Option(["-comp_based_stats", "comp_based_stats"],
                    r"""Use composition-based statistics.

                    D or d: default (equivalent to 0 )

                    0 or F or f: Simplified Composition-based statistics as in
                    Bioinformatics 15:1000-1011, 1999

                    1 or T or t: Composition-based statistics as in NAR 29:2994-3005, 2001

                    Default = 0.
                    """,
                    checker_function=lambda value: value in "Dd0Ff1Tt",
                    equate=False),
            # Extension options:
            _Switch(["-ungapped", "ungapped"],
                    r"Perform ungapped alignment only?"),
            # Miscellaneous options:
            _Switch(["-use_sw_tback", "use_sw_tback"],
                    r"Compute locally optimal Smith-Waterman alignments?"),
        ]
        _NcbiblastCommandline.__init__(self, cmd, **kwargs)


class NcbiblastformatterCommandline(_NcbibaseblastCommandline):
    """Wrapper for the NCBI BLAST+ program blast_formatter.

    With the release of BLAST 2.2.24+ (i.e. the BLAST suite rewritten in C++
    instead of C), the NCBI added the ASN.1 output format option to all the
    search tools, and extended the blast_formatter to support this as input.

    The blast_formatter command allows you to convert the ASN.1 output into
    the other output formats (XML, tabular, plain text, HTML).

    >>> from Bio.Blast.Applications import NcbiblastformatterCommandline
    >>> cline = NcbiblastformatterCommandline(archive="example.asn", outfmt=5, out="example.xml")
    >>> cline
    NcbiblastformatterCommandline(cmd='blast_formatter', out='example.xml', outfmt=5, archive='example.asn')
    >>> print(cline)
    blast_formatter -out example.xml -outfmt 5 -archive example.asn

    You would typically run the command line with cline() or via the Python
    subprocess module, as described in the Biopython tutorial.

    Note that this wrapper is for the version of blast_formatter from BLAST
    2.2.24+ (or later) which is when the NCBI first announced the inclusion
    this tool. There was actually an early version in BLAST 2.2.23+ (and
    possibly in older releases) but this did not have the -archive option
    (instead -rid is a mandatory argument), and is not supported by this
    wrapper.
    """

    def __init__(self, cmd="blast_formatter", **kwargs):
        """Initialize the class."""
        self.parameters = [
            # Input options
            _Option(["-rid", "rid"],
                    "BLAST Request ID (RID), not compatible with archive arg",
                    equate=False),
            _Option(["-archive", "archive"],
                    "Archive file of results, not compatible with rid arg.",
                    filename=True,
                    equate=False),
            # Restrict search or results
            _Option(["-max_target_seqs", "max_target_seqs"],
                    "Maximum number of aligned sequences to keep",
                    checker_function=lambda value: value >= 1,
                    equate=False),
        ]
        _NcbibaseblastCommandline.__init__(self, cmd, **kwargs)

    def _validate(self):
        incompatibles = {"rid": ["archive"]}
        self._validate_incompatibilities(incompatibles)
        _NcbibaseblastCommandline._validate(self)


class NcbideltablastCommandline(_Ncbiblast2SeqCommandline):
    """Create a commandline for the NCBI BLAST+ program deltablast (for proteins).

    This is a wrapper for the deltablast command line command included in
    the NCBI BLAST+ software (not present in the original BLAST).

    >>> from Bio.Blast.Applications import NcbideltablastCommandline
    >>> cline = NcbideltablastCommandline(query="rosemary.pro", db="nr",
    ...                               evalue=0.001, remote=True)
    >>> cline
    NcbideltablastCommandline(cmd='deltablast', query='rosemary.pro', db='nr', evalue=0.001, remote=True)
    >>> print(cline)
    deltablast -query rosemary.pro -db nr -evalue 0.001 -remote

    You would typically run the command line with cline() or via the Python
    subprocess module, as described in the Biopython tutorial.
    """

    def __init__(self, cmd="deltablast", **kwargs):
        """Initialize the class."""
        self.parameters = [
            # General search options:
            _Option(["-matrix", "matrix"],
                    "Scoring matrix name (default BLOSUM62)."),
            _Option(["-threshold", "threshold"],
                    "Minimum score for words to be added to the BLAST lookup table (float).",
                    equate=False),
            _Option(["-comp_based_stats", "comp_based_stats"],
                    """Use composition-based statistics (string, default 2, i.e. True).

                    0, F or f: no composition-based statistics.

                    2, T or t, D or d : Composition-based score adjustment as in
                    Bioinformatics 21:902-911, 2005, conditioned on sequence properties

                    Note that tblastn also supports values of 1 and 3.""",
                    checker_function=lambda value: value in "0Ft2TtDd",
                    equate=False),
            # Query filtering options:
            _Option(["-seg", "seg"],
                    """Filter query sequence with SEG (string).

                    Format: "yes", "window locut hicut", or "no" to disable.
                    Default is "12 2.2 2.5""",
                    equate=False),
            # Extension options:
            _Option(["-gap_trigger", "gap_trigger"],
                    "Number of bits to trigger gapping Default = 22",
                    equate=False),
            # Miscellaneous options:
            _Switch(["-use_sw_tback", "use_sw_tback"],
                    "Compute locally optimal Smith-Waterman alignments?"),
            # PSI-BLAST options
            _Option(["-num_iterations", "num_iterations"],
                    """Number of iterations to perform. (integer >=1, Default is 1)

                    Incompatible with: remote""",
                    equate=False),
            _Option(["-out_pssm", "out_pssm"],
                    "File name to store checkpoint file.",
                    filename=True,
                    equate=False),
            _Option(["-out_ascii_pssm", "out_ascii_pssm"],
                    "File name to store ASCII version of PSSM.",
                    filename=True,
                    equate=False),
            _Switch(["-save_pssm_after_last_round", "save_pssm_after_last_round"],
                    "Save PSSM after the last database search."),
            _Switch(["-save_each_pssm", "save_each_pssm"],
                    """Save PSSM after each iteration

                    File name is given in -save_pssm or -save_ascii_pssm options.
                    """),
            # PSSM engine options
            _Option(["-pseudocount", "pseudocount"],
                    "Pseudo-count value used when constructing PSSM (integer, default 0).",
                    equate=False),
            _Option(["-domain_inclusion_ethresh", "domain_inclusion_ethresh"],
                    """E-value inclusion threshold for alignments with conserved domains.

                    (float, Default is 0.05)""",
                    equate=False),
            _Option(["-inclusion_ethresh", "inclusion_ethresh"],
                    "Pairwise alignment e-value inclusion threshold (float, default 0.002).",
                    equate=False),
            # DELTA-BLAST options
            _Option(["-rpsdb", "rpsdb"],
                    "BLAST domain database name (dtring, Default = 'cdd_delta').",
                    equate=False),
            _Switch(["-show_domain_hits", "show_domain_hits"],
                    """Show domain hits?

                    Incompatible with:  remote, subject""")
        ]
        _Ncbiblast2SeqCommandline.__init__(self, cmd, **kwargs)


class NcbimakeblastdbCommandline(AbstractCommandline):
    """Wrapper for the NCBI BLAST+ program makeblastdb.

    This is a wrapper for the NCBI BLAST+ makeblastdb application
    to create BLAST databases. By default, this creates a blast database
    with the same name as the input file. The default output location
    is the same directory as the input.

    >>> from Bio.Blast.Applications import NcbimakeblastdbCommandline
    >>> cline = NcbimakeblastdbCommandline(dbtype="prot",
    ...                                    input_file="NC_005816.faa")
    >>> cline
    NcbimakeblastdbCommandline(cmd='makeblastdb', dbtype='prot', input_file='NC_005816.faa')
    >>> print(cline)
    makeblastdb -dbtype prot -in NC_005816.faa

    You would typically run the command line with cline() or via the Python
    subprocess module, as described in the Biopython tutorial.
    """

    def __init__(self, cmd="makeblastdb", **kwargs):
        """Initialize the class."""
        self.parameters = [
            # Basic input options
            _Switch(["-h", "h"],
                    "Print USAGE and DESCRIPTION; ignore other arguments."),
            _Switch(["-help", "help"],
                    "Print USAGE, DESCRIPTION and ARGUMENTS description; "
                    "ignore other arguments."),
            _Switch(["-version", "version"],
                    "Print version number;  ignore other arguments."),
            # Output configuration options
            _Option(["-out", "out"],
                    "Output file for alignment.",
                    filename=True,
                    equate=False),
            # makeblastdb specific options
            _Option(["-dbtype", "dbtype"],
                    "Molecule type of target db ('nucl' or 'prot')",
                    equate=False,
                    is_required=True,
                    checker_function=lambda x: x == "nucl" or x == "prot"),
            _Option(["-in", "input_file"],
                    "Input file/database name",
                    filename=True,
                    equate=False),
            _Option(["-input_type", "input_type"],
                    "Type of the data specified in input_file. "
                    "Default = 'fasta'. Added in BLAST 2.2.26.",
                    filename=False,
                    equate=False,
                    checker_function=self._input_type_checker),
            _Option(["-title", "title"],
                    "Title for BLAST database",
                    filename=False,
                    equate=False),
            _Switch(["-parse_seqids", "parse_seqids"],
                    "Option to parse seqid for FASTA input if set, for all "
                    "other input types seqids are parsed automatically"),
            _Switch(["-hash_index", "hash_index"],
                    "Create index of sequence hash values."),
            _Option(["-mask_data", "mask_data"],
                    "Comma-separated list of input files containing masking "
                    "data as produced by NCBI masking applications "
                    "(e.g. dustmasker, segmasker, windowmasker)",
                    filename=True,
                    equate=False),
            _Option(["-mask_id", "mask_id"],
                    "Comma-separated list of strings to uniquely identify the "
                    "masking algorithm",
                    filename=False,
                    equate=False),
            _Option(["-mask_desc", "mask_desc"],
                    "Comma-separated list of free form strings to describe "
                    "the masking algorithm details",
                    filename=False,
                    equate=False),
            _Switch(["-gi_mask", "gi_mask"],
                    "Create GI indexed masking data."),
            _Option(["-gi_mask_name", "gi_mask_name"],
                    "Comma-separated list of masking data output files.",
                    filename=False,
                    equate=False),
            _Option(["-max_file_sz", "max_file_sz"],
                    "Maximum file size for BLAST database files. "
                    "Default = '1GB'",
                    filename=False,
                    equate=False),
            _Option(["-logfile", "logfile"],
                    "File to which the program log should be redirected",
                    filename=True,
                    equate=False),
            _Option(["-taxid", "taxid"],
                    "Taxonomy ID to assign to all sequences",
                    filename=False,
                    equate=False,
                    checker_function=lambda x: type(x)(int(x)) == x),
            _Option(["-taxid_map", "taxid_map"],
                    "Text file mapping sequence IDs to taxonomy IDs. "
                    "Format:<SequenceId> <TaxonomyId><newline>",
                    filename=True,
                    equate=False),
        ]
        AbstractCommandline.__init__(self, cmd, **kwargs)

    def _input_type_checker(command, x):
        return x in ("asn1_bin", "asn1_txt", "blastdb", "fasta")

    def _validate(self):
        incompatibles = {"mask_id": ["gi_mask"],
                         "gi_mask": ["mask_id"],
                         "taxid": ["taxid_map"]}

        # Copied from _NcbibaseblastCommandline class above.
        # Code repeated here for python2 and 3 comptaibility,
        # because this is not a _NcbibaseblastCommandline subclass.
        for a in incompatibles:
            if self._get_parameter(a):
                for b in incompatibles[a]:
                    if self._get_parameter(b):
                        raise ValueError("Options %s and %s are incompatible."
                                         % (a, b))

        if self.mask_id and not self.mask_data:
            raise ValueError("Option mask_id requires mask_data to be set.")
        if self.mask_desc and not self.mask_id:
            raise ValueError("Option mask_desc requires mask_id to be set.")
        if self.gi_mask and not self.parse_seqids:
            raise ValueError("Option gi_mask requires parse_seqids to be set.")
        if self.gi_mask_name and not (self.mask_data and self.gi_mask):
            raise ValueError("Option gi_mask_name requires mask_data and "
                             "gi_mask to be set.")
        if self.taxid_map and not self.parse_seqids:
            raise ValueError("Option taxid_map requires parse_seqids "
                             "to be set.")
        AbstractCommandline._validate(self)


def _test():
    """Run the Bio.Blast.Applications module's doctests (PRIVATE)."""
    import doctest
    doctest.testmod(verbose=1)


if __name__ == "__main__":
    # Run the doctests
    _test()



# Copyright 2010-2011, 2013-2014, 2016-2018 by Peter Cock.  All rights reserved.
#
# This file is part of the Biopython distribution and governed by your
# choice of the "Biopython License Agreement" or the "BSD 3-Clause License".
# Please see the LICENSE file that should have been included as part of this
# package.
"""Provides code to access the TogoWS integrated websevices of DBCLS, Japan.

This module aims to make the TogoWS (from DBCLS, Japan) easier to use. See:
http://togows.dbcls.jp/

The TogoWS REST service provides simple access to a range of databases, acting
as a proxy to shield you from all the different provider APIs. This works using
simple URLs (which this module will construct for you). For more details, see
http://togows.dbcls.jp/site/en/rest.html

The functionality is somewhat similar to Biopython's Bio.Entrez module which
provides access to the NCBI's Entrez Utilities (E-Utils) which also covers a
wide range of databases.

Currently TogoWS does not provide any usage guidelines (unlike the NCBI whose
requirements are reasonably clear). To avoid risking overloading the service,
Biopython will only allow three calls per second.

The TogoWS SOAP service offers a more complex API for calling web services
(essentially calling remote functions) provided by DDBJ, KEGG and PDBj. For
example, this allows you to run a remote BLAST search at the DDBJ. This is
not yet covered by this module, however there are lots of Python examples
on the TogoWS website using the SOAPpy python library. See:
http://togows.dbcls.jp/site/en/soap.html
http://soapy.sourceforge.net/
"""

from __future__ import print_function

import time
from Bio._py3k import _binary_to_string_handle, _as_bytes

# Importing these functions with leading underscore as not intended for reuse
from Bio._py3k import urlopen as _urlopen
from Bio._py3k import quote as _quote


# Constant
_BASE_URL = "http://togows.dbcls.jp"

# Caches:
_search_db_names = None
_entry_db_names = None
_entry_db_fields = {}
_entry_db_formats = {}
_convert_formats = []


def _get_fields(url):
    """Query a TogoWS URL for a plain text list of values (PRIVATE)."""
    handle = _open(url)
    fields = handle.read().strip().split()
    handle.close()
    return fields


def _get_entry_dbs():
    return _get_fields(_BASE_URL + "/entry")


def _get_entry_fields(db):
    return _get_fields(_BASE_URL + "/entry/%s?fields" % db)


def _get_entry_formats(db):
    return _get_fields(_BASE_URL + "/entry/%s?formats" % db)


def _get_convert_formats():
    return [pair.split(".") for pair in _get_fields(_BASE_URL + "/convert/")]


def entry(db, id, format=None, field=None):
    """Call TogoWS 'entry' to fetch a record.

    Arguments:
     - db - database (string), see list below.
     - id - identier (string) or a list of identifiers (either as a list of
       strings or a single string with comma separators).
     - format - return data file format (string), options depend on the database
       e.g. "xml", "json", "gff", "fasta", "ttl" (RDF Turtle)
     - field - specific field from within the database record (string)
       e.g. "au" or "authors" for pubmed.

    At the time of writing, this includes the following::

        KEGG: compound, drug, enzyme, genes, glycan, orthology, reaction,
              module, pathway
        DDBj: ddbj, dad, pdb
        NCBI: nuccore, nucest, nucgss, nucleotide, protein, gene, onim,
              homologue, snp, mesh, pubmed
        EBI:  embl, uniprot, uniparc, uniref100, uniref90, uniref50

    For the current list, please see http://togows.dbcls.jp/entry/

    This function is essentially equivalent to the NCBI Entrez service
    EFetch, available in Biopython as Bio.Entrez.efetch(...), but that
    does not offer field extraction.
    """
    global _entry_db_names, _entry_db_fields, fetch_db_formats
    if _entry_db_names is None:
        _entry_db_names = _get_entry_dbs()
    if db not in _entry_db_names:
        raise ValueError(
            "TogoWS entry fetch does not officially support database '%s'." % db
        )
    if field:
        try:
            fields = _entry_db_fields[db]
        except KeyError:
            fields = _get_entry_fields(db)
            _entry_db_fields[db] = fields
        if db == "pubmed" and field == "ti" and "title" in fields:
            # Backwards compatibility fix for TogoWS change Nov/Dec 2013
            field = "title"
            import warnings

            warnings.warn(
                "TogoWS dropped 'pubmed' field alias 'ti', please use 'title' instead."
            )
        if field not in fields:
            raise ValueError(
                "TogoWS entry fetch does not explicitly support "
                "field '%s' for database '%s'. Only: %s"
                % (field, db, ", ".join(sorted(fields)))
            )
    if format:
        try:
            formats = _entry_db_formats[db]
        except KeyError:
            formats = _get_entry_formats(db)
            _entry_db_formats[db] = formats
        if format not in formats:
            raise ValueError(
                "TogoWS entry fetch does not explicitly support "
                "format '%s' for database '%s'. Only: %s"
                % (format, db, ", ".join(sorted(formats)))
            )

    if isinstance(id, list):
        id = ",".join(id)
    url = _BASE_URL + "/entry/%s/%s" % (db, _quote(id))
    if field:
        url += "/" + field
    if format:
        url += "." + format
    return _open(url)


def search_count(db, query):
    """Call TogoWS search count to see how many matches a search gives.

    Arguments:
     - db - database (string), see http://togows.dbcls.jp/search
     - query - search term (string)

    You could then use the count to download a large set of search results in
    batches using the offset and limit options to Bio.TogoWS.search(). In
    general however the Bio.TogoWS.search_iter() function is simpler to use.
    """
    global _search_db_names
    if _search_db_names is None:
        _search_db_names = _get_fields(_BASE_URL + "/search")
    if db not in _search_db_names:
        # TODO - Make this a ValueError? Right now despite the HTML website
        # claiming to, the "gene" or "ncbi-gene" don't work and are not listed.
        import warnings

        warnings.warn(
            "TogoWS search does not officially support database '%s'. "
            "See %s/search/ for options." % (db, _BASE_URL)
        )
    url = _BASE_URL + "/search/%s/%s/count" % (db, _quote(query))
    handle = _open(url)
    data = handle.read()
    handle.close()
    if not data:
        raise ValueError("TogoWS returned no data from URL %s" % url)
    try:
        return int(data.strip())
    except ValueError:
        raise ValueError("Expected an integer from URL %s, got: %r" % (url, data))


def search_iter(db, query, limit=None, batch=100):
    """Call TogoWS search iterating over the results (generator function).

    Arguments:
     - db - database (string), see http://togows.dbcls.jp/search
     - query - search term (string)
     - limit - optional upper bound on number of search results
     - batch - number of search results to pull back each time talk to
       TogoWS (currently limited to 100).

    You would use this function within a for loop, e.g.

    >>> from Bio import TogoWS
    >>> for id in TogoWS.search_iter("pubmed", "diabetes+human", limit=10):
    ...     print("PubMed ID: %s" %id) # maybe fetch data with entry?
    PubMed ID: ...

    Internally this first calls the Bio.TogoWS.search_count() and then
    uses Bio.TogoWS.search() to get the results in batches.
    """
    count = search_count(db, query)
    if not count:
        return
    # NOTE - We leave it to TogoWS to enforce any upper bound on each
    # batch, they currently return an HTTP 400 Bad Request if above 100.
    remain = count
    if limit is not None:
        remain = min(remain, limit)
    offset = 1  # They don't use zero based counting
    prev_ids = []  # Just cache the last batch for error checking
    while remain:
        batch = min(batch, remain)
        # print("%r left, asking for %r" % (remain, batch))
        ids = search(db, query, offset, batch).read().strip().split()
        assert len(ids) == batch, "Got %i, expected %i" % (len(ids), batch)
        # print("offset %i, %s ... %s" % (offset, ids[0], ids[-1]))
        if ids == prev_ids:
            raise RuntimeError("Same search results for previous offset")
        for identifier in ids:
            if identifier in prev_ids:
                raise RuntimeError("Result %s was in previous batch" % identifier)
            yield identifier
        offset += batch
        remain -= batch
        prev_ids = ids


def search(db, query, offset=None, limit=None, format=None):
    """Call TogoWS search.

    This is a low level wrapper for the TogoWS search function, which
    can return results in a several formats. In general, the search_iter
    function is more suitable for end users.

    Arguments:
     - db - database (string), see http://togows.dbcls.jp/search/
     - query - search term (string)
     - offset, limit - optional integers specifying which result to start from
       (1 based) and the number of results to return.
     - format - return data file format (string), e.g. "json", "ttl" (RDF)
       By default plain text is returned, one result per line.

    At the time of writing, TogoWS applies a default count limit of 100
    search results, and this is an upper bound. To access more results,
    use the offset argument or the search_iter(...) function.

    TogoWS supports a long list of databases, including many from the NCBI
    (e.g. "ncbi-pubmed" or "pubmed", "ncbi-genbank" or "genbank", and
    "ncbi-taxonomy"), EBI (e.g. "ebi-ebml" or "embl", "ebi-uniprot" or
    "uniprot, "ebi-go"), and KEGG (e.g. "kegg-compound" or "compound").
    For the current list, see http://togows.dbcls.jp/search/

    The NCBI provide the Entrez Search service (ESearch) which is similar,
    available in Biopython as the Bio.Entrez.esearch() function.

    See also the function Bio.TogoWS.search_count() which returns the number
    of matches found, and the Bio.TogoWS.search_iter() function which allows
    you to iterate over the search results (taking care of batching for you).
    """
    global _search_db_names
    if _search_db_names is None:
        _search_db_names = _get_fields(_BASE_URL + "/search")
    if db not in _search_db_names:
        # TODO - Make this a ValueError? Right now despite the HTML website
        # claiming to, the "gene" or "ncbi-gene" don't work and are not listed.
        import warnings

        warnings.warn(
            "TogoWS search does not explicitly support database '%s'. "
            "See %s/search/ for options." % (db, _BASE_URL)
        )
    url = _BASE_URL + "/search/%s/%s" % (db, _quote(query))
    if offset is not None and limit is not None:
        try:
            offset = int(offset)
        except ValueError:
            raise ValueError(
                "Offset should be an integer (at least one), not %r" % offset
            )
        try:
            limit = int(limit)
        except ValueError:
            raise ValueError(
                "Limit should be an integer (at least one), not %r" % limit
            )
        if offset <= 0:
            raise ValueError("Offset should be at least one, not %i" % offset)
        if limit <= 0:
            raise ValueError("Count should be at least one, not %i" % limit)
        url += "/%i,%i" % (offset, limit)
    elif offset is not None or limit is not None:
        raise ValueError("Expect BOTH offset AND limit to be provided (or neither)")
    if format:
        url += "." + format
    # print(url)
    return _open(url)


def convert(data, in_format, out_format):
    """Call TogoWS for file format conversion.

    Arguments:
     - data - string or handle containing input record(s)
     - in_format - string describing the input file format (e.g. "genbank")
     - out_format - string describing the requested output format (e.g. "fasta")

    For a list of supported conversions (e.g. "genbank" to "fasta"), see
    http://togows.dbcls.jp/convert/

    Note that Biopython has built in support for conversion of sequence and
    alignnent file formats (functions Bio.SeqIO.convert and Bio.AlignIO.convert)
    """
    global _convert_formats
    if not _convert_formats:
        _convert_formats = _get_convert_formats()
    if [in_format, out_format] not in _convert_formats:
        msg = "\n".join("%s -> %s" % tuple(pair) for pair in _convert_formats)
        raise ValueError("Unsupported conversion. Choose from:\n%s" % msg)
    url = _BASE_URL + "/convert/%s.%s" % (in_format, out_format)
    # TODO - Should we just accept a string not a handle? What about a filename?
    if hasattr(data, "read"):
        # Handle
        return _open(url, post=data.read())
    else:
        # String
        return _open(url, post=data)


def _open(url, post=None):
    """Build the URL and open a handle to it (PRIVATE).

    Open a handle to TogoWS, will raise an IOError if it encounters an error.

    In the absence of clear guidelines, this function enforces a limit of
    "up to three queries per second" to avoid abusing the TogoWS servers.
    """
    delay = 0.333333333  # one third of a second
    current = time.time()
    wait = _open.previous + delay - current
    if wait > 0:
        time.sleep(wait)
        _open.previous = current + wait
    else:
        _open.previous = current

    # print(url)
    if post:
        handle = _urlopen(url, _as_bytes(post))
    else:
        handle = _urlopen(url)

    # We now trust TogoWS to have set an HTTP error code, that
    # suffices for my current unit tests. Previously we would
    # examine the start of the data returned back.
    return _binary_to_string_handle(handle)


_open.previous = 0


if __name__ == "__main__":
    from Bio._utils import run_doctest

    run_doctest(verbose=0)



# Copyright 2004-2008 by M de Hoon.
# All rights reserved.
# This code is part of the Biopython distribution and governed by its
# license.  Please see the LICENSE file that should have been included
# as part of this package.
"""Implements the Lowess function for nonparametric regression.

Functions:
lowess        Fit a smooth nonparametric regression curve to a scatterplot.

For more information, see

William S. Cleveland: "Robust locally weighted regression and smoothing
scatterplots", Journal of the American Statistical Association, December 1979,
volume 74, number 368, pp. 829-836.

William S. Cleveland and Susan J. Devlin: "Locally weighted regression: An
approach to regression analysis by local fitting", Journal of the American
Statistical Association, September 1988, volume 83, number 403, pp. 596-610.
"""

from __future__ import print_function

from Bio._py3k import range

import numpy


try:
    from Bio.Cluster import median

    # The function median in Bio.Cluster is faster than the function median
    # in NumPy, as it does not require a full sort.
except ImportError as x:
    # Use the median function in NumPy if Bio.Cluster is not available
    from numpy import median


def lowess(x, y, f=2.0 / 3.0, iter=3):
    """Lowess smoother: Robust locally weighted regression.

    The lowess function fits a nonparametric regression curve to a scatterplot.
    The arrays x and y contain an equal number of elements; each pair
    (x[i], y[i]) defines a data point in the scatterplot. The function returns
    the estimated (smooth) values of y.

    The smoothing span is given by f. A larger value for f will result in a
    smoother curve. The number of robustifying iterations is given by iter. The
    function will run faster with a smaller number of iterations.

    x and y should be numpy float arrays of equal length.  The return value is
    also a numpy float array of that length.

    e.g.

    >>> import numpy
    >>> x = numpy.array([4,  4,  7,  7,  8,  9, 10, 10, 10, 11, 11, 12, 12, 12,
    ...                 12, 13, 13, 13, 13, 14, 14, 14, 14, 15, 15, 15, 16, 16,
    ...                 17, 17, 17, 18, 18, 18, 18, 19, 19, 19, 20, 20, 20, 20,
    ...                 20, 22, 23, 24, 24, 24, 24, 25], numpy.float)
    >>> y = numpy.array([2, 10,  4, 22, 16, 10, 18, 26, 34, 17, 28, 14, 20, 24,
    ...                 28, 26, 34, 34, 46, 26, 36, 60, 80, 20, 26, 54, 32, 40,
    ...                 32, 40, 50, 42, 56, 76, 84, 36, 46, 68, 32, 48, 52, 56,
    ...                 64, 66, 54, 70, 92, 93, 120, 85], numpy.float)
    >>> result = lowess(x, y)
    >>> len(result)
    50
    >>> print("[%0.2f, ..., %0.2f]" % (result[0], result[-1]))
    [4.85, ..., 84.98]
    """
    n = len(x)
    r = int(numpy.ceil(f * n))
    h = [numpy.sort(abs(x - x[i]))[r] for i in range(n)]
    w = numpy.clip(abs(([x] - numpy.transpose([x])) / h), 0.0, 1.0)
    w = 1 - w * w * w
    w = w * w * w
    yest = numpy.zeros(n)
    delta = numpy.ones(n)
    for iteration in range(iter):
        for i in range(n):
            weights = delta * w[:, i]
            weights_mul_x = weights * x
            b1 = numpy.dot(weights, y)
            b2 = numpy.dot(weights_mul_x, y)
            A11 = sum(weights)
            A12 = sum(weights_mul_x)
            A21 = A12
            A22 = numpy.dot(weights_mul_x, x)
            determinant = A11 * A22 - A12 * A21
            beta1 = (A22 * b1 - A12 * b2) / determinant
            beta2 = (A11 * b2 - A21 * b1) / determinant
            yest[i] = beta1 + beta2 * x[i]
        residuals = y - yest
        s = median(abs(residuals))
        delta[:] = numpy.clip(residuals / (6 * s), -1, 1)
        delta[:] = 1 - delta * delta
        delta[:] = delta * delta
    return yest


if __name__ == "__main__":
    from Bio._utils import run_doctest

    run_doctest()



# Copyright 2015 by Gert Hulselmans.  All rights reserved.
# This file is part of the Biopython distribution and governed by your
# choice of the "Biopython License Agreement" or the "BSD 3-Clause License".
# Please see the LICENSE file that should have been included as part of this
# package.

"""Parse Cluster Buster position frequency matrix files."""

from Bio import motifs


class Record(list):
    """Class to store the information in a Cluster Buster matrix table.

    The record inherits from a list containing the individual motifs.
    """

    def __str__(self):
        return "\n".join(str(motif) for motif in self)


def read(handle):
    """Read motifs in Cluster Buster position frequency matrix format from a file handle.

    Cluster Buster motif format: http://zlab.bu.edu/cluster-buster/help/cis-format.html
    """
    motif_nbr = 0
    record = Record()
    nucleotide_counts = {"A": [], "C": [], "G": [], "T": []}
    motif_name = ""

    for line in handle:
        line = line.strip()
        if line:
            if line.startswith(">"):

                if motif_nbr != 0:
                    motif = motifs.Motif(alphabet="GATC", counts=nucleotide_counts)
                    motif.name = motif_name
                    record.append(motif)

                motif_name = line[1:].strip()
                nucleotide_counts = {"A": [], "C": [], "G": [], "T": []}
                motif_nbr += 1
            else:
                if line.startswith("#"):
                    continue

                matrix_columns = line.split()

                if len(matrix_columns) == 4:
                    [nucleotide_counts[nucleotide].append(float(nucleotide_count))
                     for nucleotide, nucleotide_count in zip(["A", "C", "G", "T"], matrix_columns)]

    motif = motifs.Motif(alphabet="GATC", counts=nucleotide_counts)
    motif.name = motif_name
    record.append(motif)

    return record


def write(motifs):
    """Return the representation of motifs in Cluster Buster position frequency matrix format."""
    lines = []
    for m in motifs:
        line = ">{0}\n".format(m.name)
        lines.append(line)
        for ACGT_counts in zip(m.counts["A"], m.counts["C"], m.counts["G"], m.counts["T"]):
            lines.append("{0:0.0f}\t{1:0.0f}\t{2:0.0f}\t{3:0.0f}\n".format(*ACGT_counts))

    # Finished; glue the lines together.
    text = "".join(lines)

    return text



# Copyright 2013 by Leighton Pritchard.  All rights reserved.
#
# This file is part of the Biopython distribution and governed by your
# choice of the "Biopython License Agreement" or the "BSD 3-Clause License".
# Please see the LICENSE file that should have been included as part of this
# package.

"""Classes to represent a KGML Pathway Map.

The KGML definition is as of release KGML v0.7.1
(http://www.kegg.jp/kegg/xml/docs/)

Classes:
 - Pathway - Specifies graph information for the pathway map
 - Relation - Specifies a relationship between two proteins or KOs,
   or protein and compound. There is an implied direction to the
   relationship in some cases.
 - Reaction - A specific chemical reaction between a substrate and
   a product.
 - Entry - A node in the pathway graph
 - Graphics - Entry subelement describing its visual representation

"""

import time
from itertools import chain
from xml.dom import minidom
import xml.etree.ElementTree as ET

from Bio._py3k import _is_int_or_long, _as_string


# Pathway
class Pathway(object):
    """Represents a KGML pathway from KEGG.

    Specifies graph information for the pathway map, as described in
    release KGML v0.7.1 (http://www.kegg.jp/kegg/xml/docs/)

    Attributes:
     - name - KEGGID of the pathway map
     - org - ko/ec/[org prefix]
     - number - map number (integer)
     - title - the map title
     - image - URL of the image map for the pathway
     - link - URL of information about the pathway
     - entries - Dictionary of entries in the pathway, keyed by node ID
     - reactions - Set of reactions in the pathway

    The name attribute has a restricted format, so we make it a property and
    enforce the formatting.

    The Pathway object is the only allowed route for adding/removing
    Entry, Reaction, or Relation elements.

    Entries are held in a dictionary and keyed by the node ID for the
    pathway graph - this allows for ready access via the Reaction/Relation
    etc. elements.  Entries must be added before reference by any other
    element.

    Reactions are held in a dictionary, keyed by node ID for the path.
    The elements referred to in the reaction must be added before the
    reaction itself.

    """

    def __init__(self):
        """Initialize the class."""
        self._name = ""
        self.org = ""
        self._number = None
        self.title = ""
        self.image = ""
        self.link = ""
        self.entries = {}
        self._reactions = {}
        self._relations = set()

    def get_KGML(self):
        """Return the pathway as a string in prettified KGML format."""
        header = "\n".join(
            [
                '<?xml version="1.0"?>',
                "<!DOCTYPE pathway SYSTEM "
                '"http://www.genome.jp/kegg/xml/'
                'KGML_v0.7.1_.dtd">',
                "<!-- Created by KGML_Pathway.py %s -->" % time.asctime(),
            ]
        )
        rough_xml = header + _as_string(ET.tostring(self.element, "utf-8"))
        reparsed = minidom.parseString(rough_xml)
        return reparsed.toprettyxml(indent="  ")

    def add_entry(self, entry):
        """Add an Entry element to the pathway."""
        # We insist that the node ID is an integer
        if not _is_int_or_long(entry.id):
            raise TypeError(
                "Node ID must be an integer, got %s (%s)" % (type(entry.id), entry.id)
            )
        entry._pathway = self  # Let the entry know about the pathway
        self.entries[entry.id] = entry

    def remove_entry(self, entry):
        """Remove an Entry element from the pathway."""
        if not _is_int_or_long(entry.id):
            raise TypeError(
                "Node ID must be an integer, got %s (%s)" % (type(entry.id), entry.id)
            )
        # We need to remove the entry from any other elements that may
        # contain it, which means removing those elements
        # TODO
        del self.entries[entry.id]

    def add_reaction(self, reaction):
        """Add a Reaction element to the pathway."""
        # We insist that the node ID is an integer and corresponds to an entry
        if not _is_int_or_long(reaction.id):
            raise ValueError(
                "Node ID must be an integer, got %s (%s)"
                % (type(reaction.id), reaction.id)
            )
        if reaction.id not in self.entries:
            raise ValueError(
                "Reaction ID %d has no corresponding entry" % reaction.id
            )
        reaction._pathway = self  # Let the reaction know about the pathway
        self._reactions[reaction.id] = reaction

    def remove_reaction(self, reaction):
        """Remove a Reaction element from the pathway."""
        if not _is_int_or_long(reaction.id):
            raise TypeError(
                "Node ID must be an integer, got %s (%s)"
                % (type(reaction.id), reaction.id)
            )
        # We need to remove the reaction from any other elements that may
        # contain it, which means removing those elements
        # TODO
        del self._reactions[reaction.id]

    def add_relation(self, relation):
        """Add a Relation element to the pathway."""
        relation._pathway = self  # Let the reaction know about the pathway
        self._relations.add(relation)

    def remove_relation(self, relation):
        """Remove a Relation element from the pathway."""
        self._relations.remove(relation)

    def __str__(self):
        """Return a readable summary description string."""
        outstr = [
            "Pathway: %s" % self.title,
            "KEGG ID: %s" % self.name,
            "Image file: %s" % self.image,
            "Organism: %s" % self.org,
            "Entries: %d" % len(self.entries),
            "Entry types:",
        ]
        for t in ["ortholog", "enzyme", "reaction", "gene", "group", "compound", "map"]:
            etype = [e for e in self.entries.values() if e.type == t]
            if len(etype):
                outstr.append("\t%s: %d" % (t, len(etype)))
        return "\n".join(outstr) + "\n"

    # Assert correct formatting of the pathway name, and other attributes
    def _getname(self):
        return self._name

    def _setname(self, value):
        if not value.startswith("path:"):
            raise ValueError(
                "Pathway name should begin with 'path:', got %s" % value
            )
        self._name = value

    def _delname(self):
        del self._name

    name = property(_getname, _setname, _delname, "The KEGGID for the pathway map.")

    def _getnumber(self):
        return self._number

    def _setnumber(self, value):
        self._number = int(value)

    def _delnumber(self):
        del self._number

    number = property(_getnumber, _setnumber, _delnumber, "The KEGG map number.")

    @property
    def compounds(self):
        """Get a list of entries of type compound."""
        return [e for e in self.entries.values() if e.type == "compound"]

    @property
    def maps(self):
        """Get a list of entries of type map."""
        return [e for e in self.entries.values() if e.type == "map"]

    @property
    def orthologs(self):
        """Get a list of entries of type ortholog."""
        return [e for e in self.entries.values() if e.type == "ortholog"]

    @property
    def genes(self):
        """Get a list of entries of type gene."""
        return [e for e in self.entries.values() if e.type == "gene"]

    @property
    def reactions(self):
        """Get a list of reactions in the pathway."""
        return self._reactions.values()

    @property
    def reaction_entries(self):
        """List of entries corresponding to each reaction in the pathway."""
        return [self.entries[i] for i in self._reactions]

    @property
    def relations(self):
        """Get a list of relations in the pathway."""
        return list(self._relations)

    @property
    def element(self):
        """Return the Pathway as a valid KGML element."""
        # The root is this Pathway element
        pathway = ET.Element("pathway")
        pathway.attrib = {
            "name": self._name,
            "org": self.org,
            "number": str(self._number),
            "title": self.title,
            "image": self.image,
            "link": self.link,
        }
        # We add the Entries in node ID order
        for eid, entry in sorted(self.entries.items()):
            pathway.append(entry.element)
        # Next we add Relations
        for relation in self._relations:
            pathway.append(relation.element)
        for eid, reaction in sorted(self._reactions.items()):
            pathway.append(reaction.element)
        return pathway

    @property
    def bounds(self):
        """Coordinate bounds for all Graphics elements in the Pathway.

        Returns the [(xmin, ymin), (xmax, ymax)] coordinates for all
        Graphics elements in the Pathway
        """
        xlist, ylist = [], []
        for b in [g.bounds for g in self.entries.values()]:
            xlist.extend([b[0][0], b[1][0]])
            ylist.extend([b[0][1], b[1][1]])
        return [(min(xlist), min(ylist)), (max(xlist), max(ylist))]


# Entry
class Entry(object):
    """Represent an Entry from KGML.

    Each Entry element is a node in the pathway graph, as described in
    release KGML v0.7.1 (http://www.kegg.jp/kegg/xml/docs/)

    Attributes:
     - id - The ID of the entry in the pathway map (integer)
     - names - List of KEGG IDs for the entry
     - type - The type of the entry
     - link - URL of information about the entry
     - reaction - List of KEGG IDs of the corresponding reactions
       (integer)
     - graphics -    List of Graphics objects describing the Entry's visual
       representation
     - components - List of component node ID for this Entry ('group')
     - alt - List of alternate names for the Entry

    NOTE: The alt attribute represents a subelement of the substrate and
    product elements in the KGML file

    """

    def __init__(self):
        """Initialize the class."""
        self._id = None
        self._names = []
        self.type = ""
        self.image = ""
        self.link = ""
        self.graphics = []
        self.components = set()
        self.alt = []
        self._pathway = None
        self._reactions = []

    def __str__(self):
        """Return readable descriptive string."""
        outstr = [
            "Entry node ID: %d" % self.id,
            "Names: %s" % self.name,
            "Type: %s" % self.type,
            "Components: %s" % self.components,
            "Reactions: %s" % self.reaction,
            "Graphics elements: %d %s" % (len(self.graphics), self.graphics),
        ]
        return "\n".join(outstr) + "\n"

    def add_component(self, element):
        """Add an element to the entry.

        If the Entry is already part of a pathway, make sure
        the component already exists.
        """
        if self._pathway is not None:
            if element.id not in self._pathway.entries:
                raise ValueError(
                    "Component %s is not an entry in the pathway" % element.id
                )
        self.components.add(element)

    def remove_component(self, value):
        """Remove the entry with the passed ID from the group."""
        self.components.remove(value)

    def add_graphics(self, entry):
        """Add the Graphics entry."""
        self.graphics.append(entry)

    def remove_graphics(self, entry):
        """Remove the Graphics entry with the passed ID from the group."""
        self.graphics.remove(entry)

    # Names may be given as a space-separated list of KEGG identifiers
    def _getname(self):
        return " ".join(self._names)

    def _setname(self, value):
        self._names = value.split()

    def _delname(self):
        self._names = []

    name = property(
        _getname, _setname, _delname, "List of KEGG identifiers for the Entry."
    )

    # Reactions may be given as a space-separated list of KEGG identifiers
    def _getreaction(self):
        return " ".join(self._reactions)

    def _setreaction(self, value):
        self._reactions = value.split()

    def _delreaction(self):
        self._reactions = []

    reaction = property(
        _getreaction,
        _setreaction,
        _delreaction,
        "List of reaction KEGG IDs for this Entry.",
    )

    # We make sure that the node ID is an integer
    def _getid(self):
        return self._id

    def _setid(self, value):
        self._id = int(value)

    def _delid(self):
        del self._id

    id = property(_getid, _setid, _delid, "The pathway graph node ID for the Entry.")

    @property
    def element(self):
        """Return the Entry as a valid KGML element."""
        # The root is this Entry element
        entry = ET.Element("entry")
        entry.attrib = {
            "id": str(self._id),
            "name": self.name,
            "link": self.link,
            "type": self.type,
        }
        if len(self._reactions):
            entry.attrib["reaction"] = self.reaction
        if len(self.graphics):
            for g in self.graphics:
                entry.append(g.element)
        if len(self.components):
            for c in self.components:
                entry.append(c.element)
        return entry

    @property
    def bounds(self):
        """Coordinate bounds for all Graphics elements in the Entry.

        Return the [(xmin, ymin), (xmax, ymax)] co-ordinates for the Entry
        Graphics elements.
        """
        xlist, ylist = [], []
        for b in [g.bounds for g in self.graphics]:
            xlist.extend([b[0][0], b[1][0]])
            ylist.extend([b[0][1], b[1][1]])
        return [(min(xlist), min(ylist)), (max(xlist), max(ylist))]

    @property
    def is_reactant(self):
        """Return true if this Entry participates in any reaction in its parent pathway."""
        for rxn in self._pathway.reactions:
            if self._id in rxn.reactant_ids:
                return True
        return False


# Component
class Component(object):
    """An Entry subelement used to represents a complex node.

    A subelement of the Entry element, used when the Entry is a complex
    node, as described in release KGML v0.7.1
    (http://www.kegg.jp/kegg/xml/docs/)

    The Component acts as a collection (with type 'group', and typically
    its own Graphics subelement), having only an ID.
    """

    def __init__(self, parent):
        """Initialize the class."""
        self._id = None
        self._parent = parent

    # We make sure that the node ID is an integer
    def _getid(self):
        return self._id

    def _setid(self, value):
        self._id = int(value)

    def _delid(self):
        del self._id

    id = property(_getid, _setid, _delid, "The pathway graph node ID for the Entry")

    @property
    def element(self):
        """Return the Component as a valid KGML element."""
        # The root is this Component element
        component = ET.Element("component")
        component.attrib = {"id": str(self._id)}
        return component


# Graphics
class Graphics(object):
    """An Entry subelement used to represents the visual representation.

    A subelement of Entry, specifying its visual representation, as
    described in release KGML v0.7.1 (http://www.kegg.jp/kegg/xml/docs/)

    Attributes:
     - name         Label for the graphics object
     - x            X-axis position of the object (int)
     - y            Y-axis position of the object (int)
     - coords       polyline co-ordinates, list of (int, int) tuples
     - type         object shape
     - width        object width (int)
     - height       object height (int)
     - fgcolor      object foreground color (hex RGB)
     - bgcolor      object background color (hex RGB)

    Some attributes are present only for specific graphics types.  For
    example, line types do not (typically) have a width.
    We permit non-DTD attributes and attribute settings, such as

    dash         List of ints, describing an on/off pattern for dashes

    """

    def __init__(self, parent):
        """Initialize the class."""
        self.name = ""
        self._x = None
        self._y = None
        self._coords = None
        self.type = ""
        self._width = None
        self._height = None
        self.fgcolor = ""
        self.bgcolor = ""
        self._parent = parent

    # We make sure that the XY coordinates, width and height are numbers
    def _getx(self):
        return self._x

    def _setx(self, value):
        self._x = float(value)

    def _delx(self):
        del self._x

    x = property(_getx, _setx, _delx, "The X coordinate for the graphics element.")

    def _gety(self):
        return self._y

    def _sety(self, value):
        self._y = float(value)

    def _dely(self):
        del self._y

    y = property(_gety, _sety, _dely, "The Y coordinate for the graphics element.")

    def _getwidth(self):
        return self._width

    def _setwidth(self, value):
        self._width = float(value)

    def _delwidth(self):
        del self._width

    width = property(
        _getwidth, _setwidth, _delwidth, "The width of the graphics element."
    )

    def _getheight(self):
        return self._height

    def _setheight(self, value):
        self._height = float(value)

    def _delheight(self):
        del self._height

    height = property(
        _getheight, _setheight, _delheight, "The height of the graphics element."
    )

    # We make sure that the polyline co-ordinates are integers, too
    def _getcoords(self):
        return self._coords

    def _setcoords(self, value):
        clist = [int(e) for e in value.split(",")]
        self._coords = [tuple(clist[i : i + 2]) for i in range(0, len(clist), 2)]

    def _delcoords(self):
        del self._coords

    coords = property(
        _getcoords,
        _setcoords,
        _delcoords,
        "Polyline coordinates for the graphics element.",
    )

    # Set default colors
    def _getfgcolor(self):
        return self._fgcolor

    def _setfgcolor(self, value):
        if value == "none":
            self._fgcolor = "#000000"  # this default defined in KGML spec
        else:
            self._fgcolor = value

    def _delfgcolor(self):
        del self._fgcolor

    fgcolor = property(_getfgcolor, _setfgcolor, _delfgcolor, "Foreground color.")

    def _getbgcolor(self):
        return self._bgcolor

    def _setbgcolor(self, value):
        if value == "none":
            self._bgcolor = "#000000"  # this default defined in KGML spec
        else:
            self._bgcolor = value

    def _delbgcolor(self):
        del self._bgcolor

    bgcolor = property(_getbgcolor, _setbgcolor, _delbgcolor, "Background color.")

    @property
    def element(self):
        """Return the Graphics as a valid KGML element."""
        # The root is this Component element
        graphics = ET.Element("graphics")
        if isinstance(self.fgcolor, str):  # Assumes that string is hexstring
            fghex = self.fgcolor
        else:  # Assumes ReportLab Color object
            fghex = "#" + self.fgcolor.hexval()[2:]
        if isinstance(self.bgcolor, str):  # Assumes that string is hexstring
            bghex = self.bgcolor
        else:  # Assumes ReportLab Color object
            bghex = "#" + self.bgcolor.hexval()[2:]
        graphics.attrib = {
            "name": self.name,
            "type": self.type,
            "fgcolor": fghex,
            "bgcolor": bghex,
        }
        for (n, attr) in [
            ("x", "_x"),
            ("y", "_y"),
            ("width", "_width"),
            ("height", "_height"),
        ]:
            if getattr(self, attr) is not None:
                graphics.attrib[n] = str(getattr(self, attr))
        if self.type == "line":  # Need to write polycoords
            graphics.attrib["coords"] = ",".join(
                [str(e) for e in chain.from_iterable(self.coords)]
            )
        return graphics

    @property
    def bounds(self):
        """Coordinate bounds for the Graphics element.

        Return the bounds of the Graphics object as an [(xmin, ymin),
        (xmax, ymax)] tuple.  Co-ordinates give the centre of the
        circle, rectangle, roundrectangle elements, so we have to
        adjust for the relevant width/height.
        """
        if self.type == "line":
            xlist = [x for x, y in self.coords]
            ylist = [y for x, y in self.coords]
            return [(min(xlist), min(ylist)), (max(xlist), max(ylist))]
        else:
            return [
                (self.x - self.width * 0.5, self.y - self.height * 0.5),
                (self.x + self.width * 0.5, self.y + self.height * 0.5),
            ]

    @property
    def centre(self):
        """Return the centre of the Graphics object as an (x, y) tuple."""
        return (
            0.5 * (self.bounds[0][0] + self.bounds[1][0]),
            0.5 * (self.bounds[0][1] + self.bounds[1][1]),
        )


# Reaction
class Reaction(object):
    """A specific chemical reaction with substrates and products.

    This describes a specific chemical reaction between one or more
    substrates and one or more products.

    Attributes:
     - id             Pathway graph node ID of the entry
     - names          List of KEGG identifier(s) from the REACTION database
     - type           String: reversible or irreversible
     - substrate      Entry object of the substrate
     - product        Entry object of the product

    """

    def __init__(self):
        """Initialize the class."""
        self._id = None
        self._names = []
        self.type = ""
        self._substrates = set()
        self._products = set()
        self._pathway = None

    def __str__(self):
        """Return an informative human-readable string."""
        outstr = [
            "Reaction node ID: %s" % self.id,
            "Reaction KEGG IDs: %s" % self.name,
            "Type: %s" % self.type,
            "Substrates: %s" % ",".join([s.name for s in self.substrates]),
            "Products: %s" % ",".join([s.name for s in self.products]),
        ]
        return "\n".join(outstr) + "\n"

    def add_substrate(self, substrate_id):
        """Add a substrate, identified by its node ID, to the reaction."""
        if self._pathway is not None:
            if int(substrate_id) not in self._pathway.entries:
                raise ValueError(
                    "Couldn't add substrate, no node ID %d in Pathway"
                    % int(substrate_id)
                )
        self._substrates.add(substrate_id)

    def add_product(self, product_id):
        """Add a product, identified by its node ID, to the reaction."""
        if self._pathway is not None:
            if int(product_id) not in self._pathway.entries:
                raise ValueError(
                    "Couldn't add product, no node ID %d in Pathway" % product_id
                )
        self._products.add(int(product_id))

    # The node ID is also the node ID of the Entry that corresponds to the
    # reaction; we get the corresponding Entry when there is an associated
    # Pathway
    def _getid(self):
        return self._id

    def _setid(self, value):
        self._id = int(value)

    def _delid(self):
        del self._id

    id = property(_getid, _setid, _delid, "Node ID for the reaction.")

    # Names may show up as a space-separated list of several KEGG identifiers
    def _getnames(self):
        return " ".join(self._names)

    def _setnames(self, value):
        self._names.extend(value.split())

    def _delnames(self):
        del self.names

    name = property(
        _getnames, _setnames, _delnames, "List of KEGG identifiers for the reaction."
    )

    # products and substrates are read-only properties, returning lists
    # of Entry objects
    @property
    def substrates(self):
        """Return list of substrate Entry elements."""
        return [self._pathway.entries[sid] for sid in self._substrates]

    @property
    def products(self):
        """Return list of product Entry elements."""
        return [self._pathway.entries[pid] for pid in self._products]

    @property
    def entry(self):
        """Return the Entry corresponding to this reaction."""
        return self._pathway.entries[self._id]

    @property
    def reactant_ids(self):
        """Return a list of substrate and product reactant IDs."""
        return self._products.union(self._substrates)

    @property
    def element(self):
        """Return KGML element describing the Reaction."""
        # The root is this Relation element
        reaction = ET.Element("reaction")
        reaction.attrib = {"id": str(self.id), "name": self.name, "type": self.type}
        for s in self._substrates:
            substrate = ET.Element("substrate")
            substrate.attrib["id"] = str(s)
            substrate.attrib["name"] = self._pathway.entries[s].name
            reaction.append(substrate)
        for p in self._products:
            product = ET.Element("product")
            product.attrib["id"] = str(p)
            product.attrib["name"] = self._pathway.entries[p].name
            reaction.append(product)
        return reaction


# Relation
class Relation(object):
    """A relationship between to products, KOs, or protein and compound.

    This describes a relationship between two products, KOs, or protein
    and compound, as described in release KGML v0.7.1
    (http://www.kegg.jp/kegg/xml/docs/)

    Attributes:
     - entry1 - The first Entry object node ID defining the
       relation (int)
     - entry2 - The second Entry object node ID defining the
       relation (int)
     - type - The relation type
     - subtypes - List of subtypes for the relation, as a list of
       (name, value) tuples

    """

    def __init__(self):
        """Initialize the class."""
        self._entry1 = None
        self._entry2 = None
        self.type = ""
        self.subtypes = []
        self._pathway = None

    def __str__(self):
        """Return a useful human-readable string."""
        outstr = [
            "Relation (subtypes: %d):" % len(self.subtypes),
            "Entry1:",
            str(self.entry1),
            "Entry2:",
            str(self.entry2),
        ]
        for s in self.subtypes:
            outstr.extend(["Subtype: %s" % s[0], str(s[1])])
        return "\n".join(outstr)

    # Properties entry1 and entry2
    def _getentry1(self):
        if self._pathway is not None:
            return self._pathway.entries[self._entry1]
        return self._entry1

    def _setentry1(self, value):
        self._entry1 = int(value)

    def _delentry1(self):
        del self._entry1

    entry1 = property(_getentry1, _setentry1, _delentry1, "Entry1 of the relation.")

    def _getentry2(self):
        if self._pathway is not None:
            return self._pathway.entries[self._entry2]
        return self._entry2

    def _setentry2(self, value):
        self._entry2 = int(value)

    def _delentry2(self):
        del self._entry2

    entry2 = property(_getentry2, _setentry2, _delentry2, "Entry2 of the relation.")

    @property
    def element(self):
        """Return KGML element describing the Relation."""
        # The root is this Relation element
        relation = ET.Element("relation")
        relation.attrib = {
            "entry1": str(self._entry1),
            "entry2": str(self._entry2),
            "type": self.type,
        }
        for (name, value) in self.subtypes:
            subtype = ET.Element("subtype")
            subtype.attrib[name] = str(value)
            relation.append(subtype)
        return relation



# Copyright 2011-2017 by Peter Cock.  All rights reserved.
#
# This file is part of the Biopython distribution and governed by your
# choice of the "Biopython License Agreement" or the "BSD 3-Clause License".
# Please see the LICENSE file that should have been included as part of this
# package.
"""Genome Diagram Feature cross-link module."""

from reportlab.lib import colors


class CrossLink(object):
    """Hold information for drawing a cross link between features."""

    def __init__(
        self, featureA, featureB, color=colors.lightgreen, border=None, flip=False
    ):
        """Create a new cross link.

        Arguments featureA and featureB should GenomeDiagram feature objects,
        or 3-tuples (track object, start, end), and currently must be on
        different tracks.

        The color and border arguments should be ReportLab colour objects, or
        for border use a boolean False for no border, otherwise it defaults to
        the same as the main colour.

        The flip argument draws an inverted cross link, useful for showing a
        mapping where one sequence has been reversed. It is conventional to
        also use a different colour (e.g. red for simple links, blue for any
        flipped links).
        """
        # Initialize attributes
        self.featureA = featureA
        self.featureB = featureB
        self.color = color  # default color to draw the feature
        self.border = border
        self.flip = flip

    @property
    def startA(self):
        """Start position of Feature A."""
        try:
            return self.featureA.start
        except AttributeError:
            track, start, end = self.featureA
            return start

    @property
    def endA(self):
        """End position of Feature A."""
        try:
            return self.featureA.end
        except AttributeError:
            track, start, end = self.featureA
            return end

    def _trackA(self, tracks):
        try:
            track, start, end = self.featureA
            assert track in tracks
            return track
        except TypeError:
            for track in tracks:
                for feature_set in track.get_sets():
                    if hasattr(feature_set, "features"):
                        if self.featureA in feature_set.features.values():
                            return track
            return None

    @property
    def startB(self):
        """Start position of Feature B."""
        try:
            return self.featureB.start
        except AttributeError:
            track, start, end = self.featureB
            return start

    @property
    def endB(self):
        """End position of Feature B."""
        try:
            return self.featureB.end
        except AttributeError:
            track, start, end = self.featureB
            return end

    def _trackB(self, tracks):
        try:
            track, start, end = self.featureB
            assert track in tracks
            return track
        except TypeError:
            for track in tracks:
                for feature_set in track.get_sets():
                    if hasattr(feature_set, "features"):
                        if self.featureB in feature_set.features.values():
                            return track
            return None



print("import: 'Bio'")
import Bio

print("import: 'Bio.Affy'")
import Bio.Affy

print("import: 'Bio.Align'")
import Bio.Align

print("import: 'Bio.Align.Applications'")
import Bio.Align.Applications

print("import: 'Bio.AlignIO'")
import Bio.AlignIO

print("import: 'Bio.Alphabet'")
import Bio.Alphabet

print("import: 'Bio.Application'")
import Bio.Application

print("import: 'Bio.Blast'")
import Bio.Blast

print("import: 'Bio.CAPS'")
import Bio.CAPS

print("import: 'Bio.Cluster'")
import Bio.Cluster

print("import: 'Bio.Compass'")
import Bio.Compass

print("import: 'Bio.Crystal'")
import Bio.Crystal

print("import: 'Bio.Data'")
import Bio.Data

print("import: 'Bio.Emboss'")
import Bio.Emboss

print("import: 'Bio.Entrez'")
import Bio.Entrez

print("import: 'Bio.ExPASy'")
import Bio.ExPASy

print("import: 'Bio.FSSP'")
import Bio.FSSP

print("import: 'Bio.GenBank'")
import Bio.GenBank

print("import: 'Bio.Geo'")
import Bio.Geo

print("import: 'Bio.HMM'")
import Bio.HMM

print("import: 'Bio.KDTree'")
import Bio.KDTree

print("import: 'Bio.KEGG'")
import Bio.KEGG

print("import: 'Bio.KEGG.Compound'")
import Bio.KEGG.Compound

print("import: 'Bio.KEGG.Enzyme'")
import Bio.KEGG.Enzyme

print("import: 'Bio.KEGG.KGML'")
import Bio.KEGG.KGML

print("import: 'Bio.KEGG.Map'")
import Bio.KEGG.Map

print("import: 'Bio.Medline'")
import Bio.Medline

print("import: 'Bio.NMR'")
import Bio.NMR

print("import: 'Bio.Nexus'")
import Bio.Nexus

print("import: 'Bio.PDB'")
import Bio.PDB

print("import: 'Bio.PDB.QCPSuperimposer'")
import Bio.PDB.QCPSuperimposer

print("import: 'Bio.Pathway'")
import Bio.Pathway

print("import: 'Bio.Pathway.Rep'")
import Bio.Pathway.Rep

print("import: 'Bio.Phylo'")
import Bio.Phylo

print("import: 'Bio.Phylo.Applications'")
import Bio.Phylo.Applications

print("import: 'Bio.Phylo.PAML'")
import Bio.Phylo.PAML

print("import: 'Bio.PopGen'")
import Bio.PopGen

print("import: 'Bio.PopGen.GenePop'")
import Bio.PopGen.GenePop

print("import: 'Bio.Restriction'")
import Bio.Restriction

print("import: 'Bio.SCOP'")
import Bio.SCOP

print("import: 'Bio.SVDSuperimposer'")
import Bio.SVDSuperimposer

print("import: 'Bio.SearchIO'")
import Bio.SearchIO

print("import: 'Bio.SearchIO.BlastIO'")
import Bio.SearchIO.BlastIO

print("import: 'Bio.SearchIO.ExonerateIO'")
import Bio.SearchIO.ExonerateIO

print("import: 'Bio.SearchIO.HmmerIO'")
import Bio.SearchIO.HmmerIO

print("import: 'Bio.SearchIO._model'")
import Bio.SearchIO._model

print("import: 'Bio.SeqIO'")
import Bio.SeqIO

print("import: 'Bio.SeqUtils'")
import Bio.SeqUtils

print("import: 'Bio.Sequencing'")
import Bio.Sequencing

print("import: 'Bio.Sequencing.Applications'")
import Bio.Sequencing.Applications

print("import: 'Bio.Statistics'")
import Bio.Statistics

print("import: 'Bio.SubsMat'")
import Bio.SubsMat

print("import: 'Bio.SwissProt'")
import Bio.SwissProt

print("import: 'Bio.TogoWS'")
import Bio.TogoWS

print("import: 'Bio.UniGene'")
import Bio.UniGene

print("import: 'Bio.UniProt'")
import Bio.UniProt

print("import: 'Bio.Wise'")
import Bio.Wise

print("import: 'Bio._py3k'")
import Bio._py3k

print("import: 'Bio.codonalign'")
import Bio.codonalign

print("import: 'Bio.cpairwise2'")
import Bio.cpairwise2

print("import: 'Bio.motifs'")
import Bio.motifs

print("import: 'Bio.motifs.applications'")
import Bio.motifs.applications

print("import: 'Bio.motifs.jaspar'")
import Bio.motifs.jaspar

print("import: 'Bio.phenotype'")
import Bio.phenotype

print("import: 'Bio.trie'")
import Bio.trie

print("import: 'BioSQL'")
import BioSQL




# Copyright 2002 by Jeffrey Chang.
# All rights reserved.
#
# This file is part of the Biopython distribution and governed by your
# choice of the "Biopython License Agreement" or the "BSD 3-Clause License".
# Please see the LICENSE file that should have been included as part of this
# package.
"""Code for doing k-nearest-neighbors classification.

k Nearest Neighbors is a supervised learning algorithm that classifies
a new observation based the classes in its surrounding neighborhood.

Glossary:
 - distance   The distance between two points in the feature space.
 - weight     The importance given to each point for classification.

Classes:
 - kNN           Holds information for a nearest neighbors classifier.


Functions:
 - train        Train a new kNN classifier.
 - calculate    Calculate the probabilities of each class, given an observation.
 - classify     Classify an observation into a class.

Weighting Functions:
 - equal_weight    Every example is given a weight of 1.

"""

import numpy


class kNN(object):
    """Holds information necessary to do nearest neighbors classification.

    Attribues:
     - classes  Set of the possible classes.
     - xs       List of the neighbors.
     - ys       List of the classes that the neighbors belong to.
     - k        Number of neighbors to look at.
    """

    def __init__(self):
        """Initialize."""
        self.classes = set()
        self.xs = []
        self.ys = []
        self.k = None


def equal_weight(x, y):
    """Return integer one (dummy method for equally weighting)."""
    # everything gets 1 vote
    return 1


def train(xs, ys, k, typecode=None):
    """Train a k nearest neighbors classifier on a training set.

    xs is a list of observations and ys is a list of the class assignments.
    Thus, xs and ys should contain the same number of elements.  k is
    the number of neighbors that should be examined when doing the
    classification.
    """
    knn = kNN()
    knn.classes = set(ys)
    knn.xs = numpy.asarray(xs, typecode)
    knn.ys = ys
    knn.k = k
    return knn


def calculate(knn, x, weight_fn=None, distance_fn=None):
    """Calculate the probability for each class.

    Arguments:
     - x is the observed data.
     - weight_fn is an optional function that takes x and a training
       example, and returns a weight.
     - distance_fn is an optional function that takes two points and
       returns the distance between them.  If distance_fn is None (the
       default), the Euclidean distance is used.

    Returns a dictionary of the class to the weight given to the class.
    """
    if weight_fn is None:
        weight_fn = equal_weight

    x = numpy.asarray(x)

    order = []  # list of (distance, index)
    if distance_fn:
        for i in range(len(knn.xs)):
            dist = distance_fn(x, knn.xs[i])
            order.append((dist, i))
    else:
        # Default: Use a fast implementation of the Euclidean distance
        temp = numpy.zeros(len(x))
        # Predefining temp allows reuse of this array, making this
        # function about twice as fast.
        for i in range(len(knn.xs)):
            temp[:] = x - knn.xs[i]
            dist = numpy.sqrt(numpy.dot(temp, temp))
            order.append((dist, i))
    order.sort()

    # first 'k' are the ones I want.
    weights = {}  # class -> number of votes
    for k in knn.classes:
        weights[k] = 0.0
    for dist, i in order[:knn.k]:
        klass = knn.ys[i]
        weights[klass] = weights[klass] + weight_fn(x, knn.xs[i])

    return weights


def classify(knn, x, weight_fn=None, distance_fn=None):
    """Classify an observation into a class.

    If not specified, weight_fn will give all neighbors equal weight.
    distance_fn is an optional function that takes two points and returns
    the distance between them.  If distance_fn is None (the default),
    the Euclidean distance is used.
    """
    if weight_fn is None:
        weight_fn = equal_weight

    weights = calculate(
        knn, x, weight_fn=weight_fn, distance_fn=distance_fn)

    most_class = None
    most_weight = None
    for klass, weight in weights.items():
        if most_class is None or weight > most_weight:
            most_class = klass
            most_weight = weight
    return most_class



# Copyright 2019, National Marrow Donor Program (NMPD).  All rights reserved.
# Written by Peter Cock, The James Hutton Institute, under contract to NMDP.
#
# This file is part of the Biopython distribution and governed by your
# choice of the "Biopython License Agreement" or the "BSD 3-Clause License".
# Please see the LICENSE file that should have been included as part of this
# package.
"""Bio.AlignIO support for GCG MSF format.

The file format was produced by the GCG PileUp and and LocalPileUp tools,
and later tools such as T-COFFEE and MUSCLE support it as an optional
output format.

The original GCG toola would write gaps at ends of each sequence which could
be missing data as tildes (``~``), whereas internal gaps were periods (``.``)
instead. This parser replaces both with minus signs (``-``) for consistency
with the rest of ``Bio.AlignIO``.

You are expected to use this module via the Bio.AlignIO functions (or the
Bio.SeqIO functions if you want to work directly with the gapped sequences).
"""

from __future__ import print_function

from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio.Align import MultipleSeqAlignment
from .Interfaces import AlignmentIterator


class MsfIterator(AlignmentIterator):
    """GCG MSF alignment iterator."""

    _header = None  # for caching lines between __next__ calls

    def __next__(self):
        """Parse the next alignment from the handle."""
        handle = self.handle

        if self._header is None:
            line = handle.readline()
        else:
            # Header we saved from when we were parsing
            # the previous alignment.
            line = self._header
            self._header = None

        if not line:
            raise StopIteration

        # Whitelisted headers we know about.
        known_headers = ["!!NA_MULTIPLE_ALIGNMENT", "!!AA_MULTIPLE_ALIGNMENT", "PileUp"]
        # Examples in "Molecular Biology Software Training Manual GCG version 10"
        # by BBSRC Bioscuences IT Services (BITS), Harpenden, UK, Copyright 1996-2001
        # would often start as follows:
        #
        # !!AA_MUTIPLE_ALIGNMENT 1.0
        # PileUp of: @/usr/users2/culhane/...
        #
        # etc with other seemingly free format text before getting to the
        # MSF/Type/Check line and the following Name: lines block and // line.
        #
        # MUSCLE just has a line "PileUp", while other sources just use the line
        # "!!AA_MULTIPLE_ALIGNMENT" (amino acid) or "!!NA_MULTIPLE_ALIGNMENT"
        # (nucleotide).
        if line.strip().split()[0] not in known_headers:
            raise ValueError(
                "%s is not a known GCG MSF header: %s"
                % (line.strip().split()[0], ", ".join(known_headers))
            )

        while line and " MSF: " not in line:
            line = handle.readline()

        if not line:
            raise ValueError("Reached end of file without MSF/Type/Check header line")

        # Quoting from "Molecular Biology Software Training Manual GCG version 10"
        # by BBSRC Bioscuences IT Services (BITS), Harpenden, UK. Copyright 1996-2001.
        # Page 31:
        #
        # "Header information is before a .. (double dot) in a GCG format file.
        #  The file will also have a checksum specific for that file."
        #
        # This was followed by a single non-aligned sequence, but this convention
        # appears to also be used in the GCG MSF files. Quoting other examples in
        # this reference, page 31:
        #
        # localpileup_17.msf  MSF: 195  Type: P  January 6, 2000 15:41  Check: 4365 ..
        #
        # Except from page 148:
        #
        # localpileup_106.msf  MSF: 457  Type: P  November 28, 2000 16:09  Check: 2396 ..
        #
        # Quoting output from MUSCLE v3.8, have two leading spaces and a zero checksum:
        #
        #   MSF: 689  Type: N  Check: 0000  ..
        #
        # By observation, the MSF value is the column count, type is N (nucleotide)
        # or P (protein / amino acid).
        #
        # In a possible bug, EMBOSS v6.6.0.0 uses CompCheck: rather than Check: as shown,
        #
        # $ seqret -sequence Tests/Fasta/f002 -auto -stdout -osformat msf
        # !!NA_MULTIPLE_ALIGNMENT 1.0
        #
        #   stdout MSF: 633 Type: N 01/08/19 CompCheck: 8543 ..
        #
        #   Name: G26680     Len: 633  Check: 4334 Weight: 1.00
        #   Name: G26685     Len: 633  Check: 3818 Weight: 1.00
        #   Name: G29385     Len: 633  Check:  391 Weight: 1.00
        #
        # //
        #
        parts = line.strip("\n").split()
        offset = parts.index("MSF:")
        if (
            parts[offset + 2] != "Type:"
            or parts[-3] not in ("Check:", "CompCheck:")
            or parts[-1] != ".."
        ):
            raise ValueError(
                "GCG MSF header line should be "
                "'<optional text> MSF: <int> Type: <letter> <optional date> Check: <int> ..', "
                " not: %r" % line
            )
        try:
            aln_length = int(parts[offset + 1])
        except ValueError:
            aln_length = -1
        if aln_length < 0:
            raise ValueError(
                "GCG MSF header line should have MDF: <int> for column count, not %r"
                % parts[offset + 1]
            )
        seq_type = parts[offset + 3]
        if seq_type not in ["P", "N"]:
            raise ValueError(
                "GCG MSF header line should have 'Type: P' (protein) "
                "or 'Type: N' (nucleotide), not 'Type: %s'" % seq_type
            )

        # There should be a blank line after that header line, then the Name: lines
        #
        # In a possible bug, T-COFFEE v12.00 adds 'oo' after the names, as shown here,
        #
        # PileUp
        #
        #
        #
        #    MSF:  628  Type: P    Check:   147   ..
        #
        #  Name: AK1H_ECOLI/1-378 oo  Len:  628  Check:  3643  Weight:  1.000
        #  Name: AKH_HAEIN/1-382 oo  Len:  628  Check:  6504  Weight:  1.000
        #
        # //
        ids = []
        lengths = []
        checks = []
        weights = []
        line = handle.readline()
        while line and line.strip() != "//":
            line = handle.readline()
            if line.strip().startswith("Name: "):
                if " Len: " in line and " Check: " in line and " Weight: " in line:
                    rest = line[line.index("Name: ") + 6 :].strip()
                    name, rest = rest.split(" Len: ")
                    length, rest = rest.split(" Check: ")
                    check, weight = rest.split(" Weight: ")
                    name = name.strip()
                    if name.endswith(" oo"):
                        # T-COFFEE oddity, ignore this
                        name = name[:-3]
                    if name in ids:
                        raise ValueError("Duplicated ID of %r" % name)
                    if " " in name:
                        raise NotImplementedError("Space in ID %r" % name)
                    ids.append(name)
                    # Expect aln_length <= int(length.strip()), see below
                    lengths.append(int(length.strip()))
                    checks.append(int(check.strip()))
                    weights.append(float(weight.strip()))
                else:
                    raise ValueError("Malformed GCG MSF name line: %r" % line)
        if not line:
            raise ValueError("End of file while looking for end of header // line.")

        if aln_length != max(lengths):
            # In broken examples from IMGTHLA was possible to continue
            # https://github.com/ANHIG/IMGTHLA/issues/201
            max_length = max(lengths)
            max_count = sum(1 for _ in lengths if _ == max_length)
            raise ValueError(
                "GCG MSF header said alignment length %i, but %s of %i sequences said Len: %s"
                % (aln_length, max_count, len(ids), max_length)
            )

        line = handle.readline()
        if not line:
            raise ValueError("End of file after // line, expected sequences.")
        if line.strip():
            raise ValueError("After // line, expected blank line before sequences.")

        # Now load the sequences
        seqs = [[] for _ in ids]  # list of empty lists
        completed_length = 0
        while completed_length < aln_length:
            # Note might have a coordinate header line (seems to be optional)
            for idx, name in enumerate(ids):
                line = handle.readline()
                if idx == 0 and not line.strip():
                    # T-COFFEE uses two blank lines between blocks, rather than one
                    while line and not line.strip():
                        line = handle.readline()
                if not line:
                    raise ValueError("End of file where expecting sequence data.")
                # print("Looking for seq for %s in line: %r" % (name, line))
                words = line.strip().split()
                # Should we use column numbers, rather than assuming no spaces in names?
                if idx == 0 and words and words[0] != name:
                    # print("Actually have a coord line")
                    # Hopefully this is a coordinate header before the first seq
                    try:
                        i = int(words[0])
                    except ValueError:
                        i = -1
                    if i != completed_length + 1:
                        raise ValueError(
                            "Expected GCG MSF coordinate line starting %i, got: %r"
                            % (completed_length + 1, line)
                        )
                    if len(words) > 1:
                        # Final block usually not full 50 chars, so expect start only.
                        if len(words) != 2:
                            i = -1
                        else:
                            try:
                                i = int(words[1])
                            except ValueError:
                                i = -1
                        if i != (
                            completed_length + 50
                            if completed_length + 50 < aln_length
                            else aln_length
                        ):
                            raise ValueError(
                                "Expected GCG MSF coordinate line %i to %i, got: %r"
                                % (
                                    completed_length + 1,
                                    completed_length + 50
                                    if completed_length + 50 < aln_length
                                    else aln_length,
                                    line,
                                )
                            )
                    line = handle.readline()
                    words = line.strip().split()
                    # print("Still looking for seq for %s in line: %r" % (name, line))
                # Dealt with any coordinate header line, should now be sequence
                if not words:
                    # Should be sequence here, but perhaps its a short one?
                    if (
                        lengths[idx] < aln_length
                        and len("".join(seqs[idx])) == lengths[idx]
                    ):
                        # Is this actually allowed in the format? Personally I would
                        # expect a line with name and a block of trailing ~ here.
                        pass
                    else:
                        raise ValueError(
                            "Expected sequence for %s, got: %r" % (name, line)
                        )
                elif words[0] == name:
                    assert len(words) > 1, line
                    # print(i, name, repr(words))
                    seqs[idx].extend(words[1:])
                else:
                    raise ValueError("Expected sequence for %r, got: %r" % (name, line))
            # TODO - check the sequence lengths thus far are consistent
            # with blocks of 50?
            completed_length += 50
            line = handle.readline()
            if line.strip():
                raise ValueError("Expected blank line, got: %r" % line)

        # Skip over any whitespace at the end...
        while True:
            line = handle.readline()
            if not line:
                # End of file, no more alignments
                break
            elif not line.strip():
                # Blank line, ignore
                pass
            elif line.strip().split()[0] in known_headers:
                # Looks like the start of another alignment:
                self._header = line
                break
            else:
                raise ValueError("Unexpected line after GCG MSF alignment: %r" % line)

        # Combine list of strings into single string, remap gaps
        seqs = ["".join(s).replace("~", "-").replace(".", "-") for s in seqs]

        # Apply any trailing padding for short sequences
        padded = False
        for idx, (length, s) in enumerate(zip(lengths, seqs)):
            if len(s) < aln_length and len(s) == length:
                padded = True
                seqs[idx] = s + "-" * (aln_length - len(s))
        if padded:
            import warnings
            from Bio import BiopythonParserWarning

            warnings.warn(
                "One of more alignment sequences were truncated and have been gap padded",
                BiopythonParserWarning,
            )

        records = (
            SeqRecord(
                Seq(s, self.alphabet),
                id=i,
                name=i,
                description=i,
                annotations={"weight": w},
            )
            for (i, s, w) in zip(ids, seqs, weights)
        )

        # This will check alignment lengths are self-consistent:
        align = MultipleSeqAlignment(records, self.alphabet)
        # Check matches the header:
        if align.get_alignment_length() != aln_length:
            raise ValueError(
                "GCG MSF headers said alignment length %i, but have %i"
                % (aln_length, align.get_alignment_length())
            )
        return align



# Copyright 2015 by Gert Hulselmans.  All rights reserved.
# This file is part of the Biopython distribution and governed by your
# choice of the "Biopython License Agreement" or the "BSD 3-Clause License".
# Please see the LICENSE file that should have been included as part of this
# package.

"""Parse various position frequency matrix format files."""

import re

from Bio import motifs


class Record(list):
    """Class to store the information in a position frequency matrix table.

    The record inherits from a list containing the individual motifs.
    """

    def __str__(self):
        return "\n".join(str(motif) for motif in self)


def read(handle, pfm_format):
    """Read motif(s) from a file in various position frequency matrix formats.

    Return the record of PFM(s).
    Call the appropriate routine based on the format passed.
    """
    # Supporting underscores here for backward compatibility
    pfm_format = pfm_format.lower().replace("_", "-")
    if pfm_format == "pfm-four-columns":
        record = _read_pfm_four_columns(handle)
        return record
    elif pfm_format == "pfm-four-rows":
        record = _read_pfm_four_rows(handle)
        return record
    else:
        raise ValueError("Unknown Position Frequency matrix format '%s'" % pfm_format)


def _read_pfm_four_columns(handle):
    """Read motifs in Cluster Buster position frequency matrix format from a file handle.

    Cluster Buster motif format: http://zlab.bu.edu/cluster-buster/help/cis-format.html

    #cisbp
    Pos A   C   G   T
    1   0.00961538461538462 0.00961538461538462 0.00961538461538462 0.971153846153846
    2   0.00961538461538462 0.00961538461538462 0.00961538461538462 0.971153846153846
    3   0.971153846153846   0.00961538461538462 0.00961538461538462 0.00961538461538462
    4   0.00961538461538462 0.00961538461538462 0.00961538461538462 0.971153846153846
    5   0.00961538461538462 0.971153846153846   0.00961538461538462 0.00961538461538462
    6   0.971153846153846   0.00961538461538462 0.00961538461538462 0.00961538461538462
    7   0.00961538461538462 0.971153846153846   0.00961538461538462 0.00961538461538462
    8   0.00961538461538462 0.00961538461538462 0.00961538461538462 0.971153846153846

    #c2h2 zfs
    Gene    ENSG00000197372
    Pos A   C   G   T
    1   0.341303    0.132427    0.117054    0.409215
    2   0.283785    0.077066    0.364552    0.274597
    3   0.491055    0.078208    0.310520    0.120217
    4   0.492621    0.076117    0.131007    0.300256
    5   0.250645    0.361464    0.176504    0.211387
    6   0.276694    0.498070    0.197793    0.027444
    7   0.056317    0.014631    0.926202    0.002850
    8   0.004470    0.007769    0.983797    0.003964
    9   0.936213    0.058787    0.002387    0.002613
    10  0.004352    0.004030    0.002418    0.989200
    11  0.013277    0.008165    0.001991    0.976567
    12  0.968132    0.002263    0.002868    0.026737
    13  0.397623    0.052017    0.350783    0.199577
    14  0.000000    0.000000    1.000000    0.000000
    15  1.000000    0.000000    0.000000    0.000000
    16  0.000000    0.000000    1.000000    0.000000
    17  0.000000    0.000000    1.000000    0.000000
    18  1.000000    0.000000    0.000000    0.000000
    19  0.000000    1.000000    0.000000    0.000000
    20  1.000000    0.000000    0.000000    0.000000

    #c2h2 zfs
    Gene    FBgn0000210
    Motif   M1734_0.90
    Pos A   C   G   T
    1   0.25    0.0833333   0.0833333   0.583333
    2   0.75    0.166667    0.0833333   0
    3   0.833333    0   0   0.166667
    4   1   0   0   0
    5   0   0.833333    0.0833333   0.0833333
    6   0.333333    0   0   0.666667
    7   0.833333    0   0   0.166667
    8   0.5 0   0.333333    0.166667
    9   0.5 0.0833333   0.166667    0.25
    10  0.333333    0.25    0.166667    0.25
    11  0.166667    0.25    0.416667    0.166667

    # flyfactorsurvey (cluster buster)
    >AbdA_Cell_FBgn0000014
    1   3   0   14
    0   0   0   18
    16  0   0   2
    18  0   0   0
    1   0   0   17
    0   0   6   12
    15  1   2   0

    # homer
    >ATGACTCATC AP-1(bZIP)/ThioMac-PU.1-ChIP-Seq(GSE21512)/Homer    6.049537    -1.782996e+03   0   9805.3,5781.0,3085.1,2715.0,0.00e+00
    0.419   0.275   0.277   0.028
    0.001   0.001   0.001   0.997
    0.010   0.002   0.965   0.023
    0.984   0.003   0.001   0.012
    0.062   0.579   0.305   0.054
    0.026   0.001   0.001   0.972
    0.043   0.943   0.001   0.012
    0.980   0.005   0.001   0.014
    0.050   0.172   0.307   0.471
    0.149   0.444   0.211   0.195

    # hocomoco
    > AHR_si
    40.51343240527031  18.259112547756697  56.41253757072521  38.77363485291994
    10.877470982533044  11.870876719950774  34.66312982331297  96.54723985087516
    21.7165707818416  43.883079837598544  20.706746561638717  67.6523201955933
    2.5465132509466635  1.3171620263517245  145.8637051322628  4.231336967110781
    0.0  150.35847450464382  1.4927836298652875  2.1074592421627525
    3.441039751299748  0.7902972158110341  149.37613720253387  0.3512432070271259
    0.0  3.441039751299748  0.7024864140542533  149.81519121131782
    0.0  0.0  153.95871737667187  0.0
    43.07922333291745  66.87558226865211  16.159862546986584  27.844049228115868

    # neph
    UW.Motif.0001   atgactca
    0.772949    0.089579    0.098612    0.038860
    0.026652    0.004653    0.025056    0.943639
    0.017663    0.023344    0.918728    0.040264
    0.919596    0.025414    0.029759    0.025231
    0.060312    0.772259    0.104968    0.062462
    0.037406    0.020643    0.006667    0.935284
    0.047316    0.899024    0.026928    0.026732
    0.948639    0.019497    0.005737    0.026128

    # tiffin
    T   A   G   C
    30  0   28  40
    0   0   0   99
    0   55  14  29
    0   99  0   0
    20  78  0   0
    0   52  7   39
    19  46  11  22
    0   60  38  0
    0   33  0   66
    73  0   25  0
    99  0   0   0
    """
    record = Record()

    motif_name = None
    motif_nbr = 0
    motif_nbr_added = 0

    default_nucleotide_order = ["A", "C", "G", "T"]
    nucleotide_order = default_nucleotide_order
    nucleotide_counts = {"A": [], "C": [], "G": [], "T": []}

    for line in handle:
        line = line.strip()

        if line:
            columns = line.split()
            nbr_columns = len(columns)

            if line.startswith("#"):
                # Skip comment lines.
                continue
            elif line.startswith(">"):
                # Parse ">AbdA_Cell_FBgn0000014" and "> AHR_si" like lines and put the part after ">" as motif name.
                if motif_nbr != 0 and motif_nbr_added != motif_nbr:
                    # Add the previous motif to the record.
                    motif = motifs.Motif(alphabet="GATC", counts=nucleotide_counts)
                    motif.name = motif_name
                    record.append(motif)
                    motif_nbr_added = motif_nbr

                # Reinitialize variables for the new motif.
                motif_name = line[1:].strip()
                nucleotide_order = default_nucleotide_order
            elif columns[0] == "Gene":
                # Parse "Gene   ENSG00000197372" like lines and put the gene name as motif name.
                if motif_nbr != 0 and motif_nbr_added != motif_nbr:
                    # Add the previous motif to the record.
                    motif = motifs.Motif(alphabet="GATC", counts=nucleotide_counts)
                    motif.name = motif_name
                    record.append(motif)
                    motif_nbr_added = motif_nbr

                # Reinitialize variables for the new motif.
                motif_name = columns[1]
                nucleotide_order = default_nucleotide_order
            elif columns[0] == "Motif":
                # Parse "Motif  M1734_0.90" like lines.
                if motif_nbr != 0 and motif_nbr_added != motif_nbr:
                    # Add the previous motif to the record.
                    motif = motifs.Motif(alphabet="GATC", counts=nucleotide_counts)
                    motif.name = motif_name
                    record.append(motif)
                    motif_nbr_added = motif_nbr

                # Reinitialize variables for the new motif.
                motif_name = columns[1]
                nucleotide_order = default_nucleotide_order
            elif columns[0] == "Pos":
                # Parse "Pos    A   C   G   T" like lines and change nucleotide order if necessary.
                if nbr_columns == 5:
                    # If the previous line was not a "Gene  ENSG00000197372" like line, a new motif starts here.
                    if motif_nbr != 0 and motif_nbr_added != motif_nbr:
                        # Add the previous motif to the record.
                        motif = motifs.Motif(alphabet="GATC", counts=nucleotide_counts)
                        motif.name = motif_name
                        record.append(motif)
                        motif_nbr_added = motif_nbr

                    nucleotide_order = default_nucleotide_order

                    if set(columns[1:]) == set(default_nucleotide_order):
                        nucleotide_order = columns[1:]
            elif columns[0] in default_nucleotide_order:
                # Parse "A  C   G   T" like lines and change nucleotide order if necessary.
                if nbr_columns == 4:
                    nucleotide_order = default_nucleotide_order
                    if set(columns) == set(default_nucleotide_order):
                        nucleotide_order = columns
            else:
                # Parse matrix columns lines and use the correct nucleotide order.
                if nbr_columns == 4:
                    matrix_columns = columns
                elif nbr_columns == 5:
                    matrix_columns = columns[1:]
                else:
                    continue

                if motif_nbr == motif_nbr_added:
                    # A new motif matrix starts here, so reinitialize variables for the new motif.
                    nucleotide_counts = {"A": [], "C": [], "G": [], "T": []}
                    motif_nbr += 1

                [nucleotide_counts[nucleotide].append(float(nucleotide_count))
                 for nucleotide, nucleotide_count in zip(nucleotide_order, matrix_columns)]
        else:
            # Empty lines can be separators between motifs.
            if motif_nbr != 0 and motif_nbr_added != motif_nbr:
                # Add the previous motif to the record.
                motif = motifs.Motif(alphabet="GATC", counts=nucleotide_counts)
                motif.name = motif_name
                record.append(motif)
                motif_nbr_added = motif_nbr

            # Reinitialize variables for the new motif.
            motif_name = None
            nucleotide_order = default_nucleotide_order
            # nucleotide_counts = {'A': [], 'C': [], 'G': [], 'T': []}

    if motif_nbr != 0 and motif_nbr_added != motif_nbr:
        motif = motifs.Motif(alphabet="GATC", counts=nucleotide_counts)
        motif.name = motif_name
        record.append(motif)

    return record


def _read_pfm_four_rows(handle):
    """Read motifs in position frequency matrix format from a file handle.

    Cluster Buster motif format: http://zlab.bu.edu/cluster-buster/help/cis-format.html

    #hdpi
    A   0   5   6   5   1   0
    C   1   1   0   0   0   4
    G   5   0   0   0   3   0
    T   0   0   0   1   2   2

    # yetfasco
    A   0.5 0.0 0.0 0.25    0.25    0.25    0.25    0.25    0.25    0.25    0.25    0.25    0.5 0.0 0.0833333334583333
    T   0.0 0.0 0.0 0.25    0.25    0.25    0.25    0.25    0.25    0.25    0.25    0.25    0.0 0.0 0.0833333334583333
    G   0.0 1.0 0.0 0.25    0.25    0.25    0.25    0.25    0.25    0.25    0.25    0.25    0.0 1.0 0.249999999875
    C   0.5 0.0 1.0 0.25    0.25    0.25    0.25    0.25    0.25    0.25    0.25    0.25    0.5 0.0 0.583333333208333

    #flyfactorsurvey ZFP finger
    A |     92    106    231    135      0      1    780     28      0    700    739     94     60    127    130
    C |    138     82    129     81    774      1      3      1      0      6     17     49    193    122    148
    G |    270    398     54    164      7    659      1    750    755     65      1     41    202    234    205
    T |    290    204    375    411      9    127      6     11     36     20     31    605    335    307    308

    # scertf pcm
    A | 9 1 1 97 1 94
    T | 80 1 97 1 1 2
    C | 9 97 1 1 1 2
    G | 2 1 1 1 97 2

    # scertf pfm
    A | 0.090 0.010 0.010 0.970 0.010 0.940
    C | 0.090 0.970 0.010 0.010 0.010 0.020
    G | 0.020 0.010 0.010 0.010 0.970 0.020
    T | 0.800 0.010 0.970 0.010 0.010 0.020

    #idmmpmm
    > abd-A
    0.218451749734889 0.0230646871686108 0.656680805938494 0.898197242841994 0.040694591728526 0.132953340402969 0.74907211028632 0.628313891834571
    0.0896076352067868 0.317338282078473 0.321580063626723 0.0461293743372216 0.0502386002120891 0.040694591728526 0.0284994697773065 0.0339342523860021
    0.455991516436904 0.0691940615058324 0.0108695652173913 0.0217391304347826 0.0284994697773065 0.0284994697773065 0.016304347826087 0.160127253446448
    0.235949098621421 0.590402969247084 0.0108695652173913 0.0339342523860021 0.880567338282079 0.797852598091198 0.206124072110286 0.17762460233298

    # JASPAR
        >MA0001.1 AGL3
        A  [ 0  3 79 40 66 48 65 11 65  0 ]
        C  [94 75  4  3  1  2  5  2  3  3 ]
        G  [ 1  0  3  4  1  0  5  3 28 88 ]
        T  [ 2 19 11 50 29 47 22 81  1  6 ]

    or::

        >MA0001.1 AGL3
        0  3 79 40 66 48 65 11 65  0
        94 75  4  3  1  2  5  2  3  3
        1  0  3  4  1  0  5  3 28 88
        2 19 11 50 29 47 22 81  1  6
    """
    record = Record()

    name_pattern = re.compile(r"^>\s*(.+)\s*")
    row_pattern_with_nucleotide_letter = re.compile(r"\s*([ACGT])\s*[[]*[|]*\s*([0-9.\s]+)\s*[]]*\s*")
    row_pattern_without_nucleotide_letter = re.compile(r"\s*([0-9.\s]+)\s*")

    motif_name = None
    nucleotide_counts = {}
    row_count = 0
    nucleotides = ["A", "C", "G", "T"]

    for line in handle:
        line = line.strip()

        name_match = name_pattern.match(line)
        row_match_with_nucleotide_letter = row_pattern_with_nucleotide_letter.match(line)
        row_match_without_nucleotide_letter = row_pattern_without_nucleotide_letter.match(line)

        if name_match:
            motif_name = name_match.group(1)
        elif row_match_with_nucleotide_letter:
            (nucleotide, counts_str) = row_match_with_nucleotide_letter.group(1, 2)
            current_nucleotide_counts = counts_str.split()
            nucleotide_counts[nucleotide] = [float(current_nucleotide_count) for current_nucleotide_count in current_nucleotide_counts]
            row_count += 1
            if row_count == 4:
                motif = motifs.Motif(alphabet="GATC", counts=nucleotide_counts)

                if motif_name:
                    motif.name = motif_name

                record.append(motif)

                motif_name = None
                nucleotide_counts = {}
                row_count = 0
        elif row_match_without_nucleotide_letter:
            current_nucleotide_counts = row_match_without_nucleotide_letter.group(1).split()
            nucleotide_counts[nucleotides[row_count]] = [float(current_nucleotide_count) for current_nucleotide_count in current_nucleotide_counts]
            row_count += 1
            if row_count == 4:
                motif = motifs.Motif(alphabet="GATC", counts=nucleotide_counts)

                if motif_name:
                    motif.name = motif_name

                record.append(motif)

                motif_name = None
                nucleotide_counts = {}
                row_count = 0

    return record


def write(motifs):
    """Return the representation of motifs in Cluster Buster position frequency matrix format."""
    lines = []
    for m in motifs:
        line = ">{0}\n".format(m.name)
        lines.append(line)
        for ACGT_counts in zip(m.counts["A"], m.counts["C"], m.counts["G"], m.counts["T"]):
            lines.append("{0:0.0f}\t{1:0.0f}\t{2:0.0f}\t{3:0.0f}\n".format(*ACGT_counts))

    # Finished; glue the lines together.
    text = "".join(lines)

    return text



# Copyright 2003 Yair Benita.  All rights reserved.
# This file is part of the Biopython distribution and governed by your
# choice of the "Biopython License Agreement" or the "BSD 3-Clause License".
# Please see the LICENSE file that should have been included as part of this
# package.
"""Calculate isoelectric points of polypeptides using methods of Bjellqvist.

pK values and the methos are taken from::

    * Bjellqvist, B.,Hughes, G.J., Pasquali, Ch., Paquet, N., Ravier, F.,
    Sanchez, J.-Ch., Frutiger, S. & Hochstrasser, D.F.
    The focusing positions of polypeptides in immobilized pH gradients can be
    predicted from their amino acid sequences. Electrophoresis 1993, 14,
    1023-1031.

    * Bjellqvist, B., Basse, B., Olsen, E. and Celis, J.E.
    Reference points for comparisons of two-dimensional maps of proteins from
    different human cell types defined in a pH scale where isoelectric points
    correlate with polypeptide compositions. Electrophoresis 1994, 15, 529-539.

I designed the algorithm according to a note by David L. Tabb, available at:
http://fields.scripps.edu/DTASelect/20010710-pI-Algorithm.pdf
"""

positive_pKs = {"Nterm": 7.5, "K": 10.0, "R": 12.0, "H": 5.98}
negative_pKs = {"Cterm": 3.55, "D": 4.05, "E": 4.45, "C": 9.0, "Y": 10.0}
pKcterminal = {"D": 4.55, "E": 4.75}
pKnterminal = {
    "A": 7.59,
    "M": 7.0,
    "S": 6.93,
    "P": 8.36,
    "T": 6.82,
    "V": 7.44,
    "E": 7.7,
}
charged_aas = ("K", "R", "H", "D", "E", "C", "Y")


class IsoelectricPoint(object):
    """A class for calculating the IEP or charge at given pH of a protein.

    Parameters
    ----------
    :protein_sequence: A ``Bio.Seq`` or string object containing a protein
                       sequence.
    :aa_content: A dictionary with amino acid letters as keys and it's
                 occurences as integers, e.g. ``{"A": 3, "C": 0, ...}``.
                 Default: ``None``. If ``None``, the dic will be calculated
                 from the given sequence.

    Methods
    -------
    :charge_at_pH(pH):  Calculates the charge of the protein for a given pH
    :pi():              Calculates the isoelectric point


    Examples
    --------
    The methods of this class can either be accessed from the class itself
    or from a ``ProtParam.ProteinAnalysis`` object (with partially different
    names):

    >>> from Bio.SeqUtils.IsoelectricPoint import IsoelectricPoint as IP
    >>> protein = IP("INGAR")
    >>> print("IEP of peptide {} is {:.2f}"
    ...       .format(protein.sequence, protein.pi()))
    IEP of peptide INGAR is 9.75
    >>> print("It's charge at pH 7 is {:.2f}"
    ...       .format(protein.charge_at_pH(7.0)))
    It's charge at pH 7 is 0.76


    >>> from Bio.SeqUtils.ProtParam import ProteinAnalysis as PA
    >>> protein = PA("PETER")
    >>> print("IEP of {}: {:.2f}".format(protein.sequence,
    ...                                  protein.isoelectric_point()))
    IEP of PETER: 4.53
    >>> print("Charge at pH 4.53: {:.2f}"
    ...       .format(protein.charge_at_pH(4.53)))
    Charge at pH 4.53: 0.00

    """

    def __init__(self, protein_sequence, aa_content=None):
        """Initialize the class."""
        self.sequence = str(protein_sequence).upper()
        if not aa_content:
            from Bio.SeqUtils.ProtParam import ProteinAnalysis as _PA

            aa_content = _PA(self.sequence).count_amino_acids()
        self.charged_aas_content = self._select_charged(aa_content)

        self.pos_pKs, self.neg_pKs = self._update_pKs_tables()

    # This function creates a dictionary with the contents of each charged aa,
    # plus Cterm and Nterm.
    def _select_charged(self, aa_content):
        charged = {}
        for aa in charged_aas:
            charged[aa] = float(aa_content[aa])
        charged["Nterm"] = 1.0
        charged["Cterm"] = 1.0
        return charged

    def _update_pKs_tables(self):
        """Update pKs tables with seq specific values for N- and C-termini."""
        pos_pKs = positive_pKs.copy()
        neg_pKs = negative_pKs.copy()
        nterm, cterm = self.sequence[0], self.sequence[-1]
        if nterm in pKnterminal:
            pos_pKs["Nterm"] = pKnterminal[nterm]
        if cterm in pKcterminal:
            neg_pKs["Cterm"] = pKcterminal[cterm]
        return pos_pKs, neg_pKs

    # This function calculates the total charge of the protein at a given pH.
    def _chargeR(self, pH):
        positive_charge = 0.0
        for aa, pK in self.pos_pKs.items():
            CR = 10 ** (pK - pH)
            partial_charge = CR / (CR + 1.0)
            positive_charge += self.charged_aas_content[aa] * partial_charge

        negative_charge = 0.0
        for aa, pK in self.neg_pKs.items():
            CR = 10 ** (pH - pK)
            partial_charge = CR / (CR + 1.0)
            negative_charge += self.charged_aas_content[aa] * partial_charge

        return positive_charge - negative_charge

    def charge_at_pH(self, pH):
        """Calculate the charge of a protein at given pH."""
        return self._chargeR(pH)

    # This is the action function, it tries different pH until the charge of
    # the protein is 0 (or close).
    def pi(self):
        """Calculate and return the isoelectric point as float."""
        # Bracket between pH1 and pH2
        pH = 7
        charge = self._chargeR(pH)
        if charge > 0.0:
            pH1 = pH
            charge1 = charge
            while charge1 > 0.0:
                pH = pH1 + 1.0
                charge = self._chargeR(pH)
                if charge > 0.0:
                    pH1 = pH
                    charge1 = charge
                else:
                    pH2 = pH
                    charge2 = charge
                    break
        else:
            pH2 = pH
            charge2 = charge
            while charge2 < 0.0:
                pH = pH2 - 1.0
                charge = self._chargeR(pH)
                if charge < 0.0:
                    pH2 = pH
                    charge2 = charge
                else:
                    pH1 = pH
                    charge1 = charge
                    break

        # Bisection
        while pH2 - pH1 > 0.0001 and charge != 0.0:
            pH = (pH1 + pH2) / 2.0
            charge = self._chargeR(pH)
            if charge > 0.0:
                pH1 = pH
                charge1 = charge
            else:
                pH2 = pH
                charge2 = charge

        return pH



# Copyright 2002-2003 Jeff Chang.  All rights reserved.
# Revisions copyright 2012 by Christian Brueffer.  All rights reserved.
# Revisions copyright 2012-2017 by Peter Cock.  All rights reserved.
# Revisions copyright 2015 by Brian Osborne.  All rights reserved.
#
# This file is part of the Biopython distribution and governed by your
# choice of the "Biopython License Agreement" or the "BSD 3-Clause License".
# Please see the LICENSE file that should have been included as part of this
# package.
#

"""Given a trie, find all occurrences of a word in the trie in a string.

Like searching a string for a substring, except that the substring is
any word in a trie.

Functions:
 - match         Find longest key in a trie matching the beginning of the string.
 - match_all     Find all keys in a trie matching the beginning of the string.
 - find          Find keys in a trie matching anywhere in a string.
 - find_words    Find keys in a trie matching whole words in a string.

This module is DEPRECATED. We encourage users to switch to alternative libraries
implementing a trie data structure, for example pygtrie.
"""

import string
import re

from Bio import BiopythonDeprecationWarning
import warnings
warnings.warn("This module has been deprecated. We encourage users to switch "
              "to alternative libraries implementing a trie data structure, "
              "for example pygtrie.", BiopythonDeprecationWarning)


def match(string, trie):
    """Find longest key, or return None.

    Find the longest key in the trie that matches the beginning of the
    string.
    """
    longest = None
    for i in range(len(string)):
        substr = string[:i + 1]
        if not trie.has_prefix(substr):
            break
        if substr in trie:
            longest = substr
    return longest


def match_all(string, trie):
    """Find and return a list of keys.

    Find all the keys in the trie that matches the beginning of the
    string.
    """
    matches = []
    for i in range(len(string)):
        substr = string[:i + 1]
        if not trie.has_prefix(substr):
            break
        if substr in trie:
            matches.append(substr)
    return matches


def find(string, trie):
    """Find all the keys in the trie that match anywhere in the string.

    Returns a list of tuples (key, start, end).
    """
    results = []
    start = 0     # index to start the search
    while start < len(string):
        # Look for a match.
        keys = match_all(string[start:], trie)
        for key in keys:
            results.append((key, start, start + len(key)))
        start += 1
    return results


DEFAULT_BOUNDARY_CHARS = string.punctuation + string.whitespace


def find_words(string, trie):
    """Find all the keys in the trie that match full words in the string.

    Find all the keys in the trie that match full words in the string.
    Word boundaries are defined as any punctuation or whitespace.

    Returns a list of tuples (key, start, end).
    """
    _boundary_re = re.compile(r"[%s]+" % re.escape(DEFAULT_BOUNDARY_CHARS))

    results = []
    start = 0     # index of word boundary
    while start < len(string):
        # Look for a match.
        keys = match_all(string[start:], trie)
        for key in keys:
            length = len(key)
            # Make sure it ends at a boundary.
            if start + length == len(string) or \
               _boundary_re.match(string[start + length]):
                results.append((key, start, start + length))
        # Move forward to the next boundary.
        m = _boundary_re.search(string, start)
        if m is None:
            break
        start = m.end()
    return results



# This file is part of the Biopython distribution and governed by your
# choice of the "Biopython License Agreement" or the "BSD 3-Clause License".
# Please see the LICENSE file that should have been included as part of this
# package.
"""
Basic statistics module.

This module is DEPRECATED.
"""
# TODO: Remove empty __init__.py once we drop Python 2 support

import warnings
from Bio import BiopythonDeprecationWarning
warnings.warn("Bio.Statistics has been deprecated, and we intend to remove it "
              "in a future release of Biopython.", BiopythonDeprecationWarning)
